# -*- coding: utf-8 -*-
from __future__ import with_statement # allow python 2.5 to work

'''
@author: Soizic Laguitton

@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''



#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

import os
import sys
import socket
import ConfigParser

from soma_workflow.errors import ConfigurationError
import soma_workflow.observer as observer


#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

LIGHT_MODE = 'light'
REMOTE_MODE = 'remote'
LOCAL_MODE = 'local'
MODES = [LIGHT_MODE,
         REMOTE_MODE,
         LOCAL_MODE]

LOCAL_SCHEDULER = 'local_basic'
DRMAA_SCHEDULER = 'drmaa'
MPI_SCHEDULER = 'mpi'
SCHEDULER_TYPES = [LOCAL_SCHEDULER,
                   DRMAA_SCHEDULER,
                   MPI_SCHEDULER]

# configuration variables ------------------------------------------------------

'''
CFG => Mandatory items
OCFG => Optional
'''
CFG_CLUSTER_ADDRESS = 'CLUSTER_ADDRESS'
CFG_SUBMITTING_MACHINES = 'SUBMITTING_MACHINES'
OCFG_DRMAA_IMPLEMENTATION = 'DRMAA_IMPLEMENTATION'
OCFG_SCHEDULER_TYPE = 'SCHEDULER_TYPE'


#OCFG_QUEUES is a list of queue name separated by white spaces.
#ex: "queue1 queue2"
OCFG_QUEUES = 'QUEUES'

OCFG_LOGIN = 'LOGIN'

OCFG_SSHPort = 'SSHPort'
OCFG_INSTALLPATH = 'INSTALLPATH'

#OCFG_MAX_JOB_IN_QUEUE allow to specify a maximum number of job N which can be
#in the queue for one user. The engine won't submit more than N job at once. The 
#also wait for the job to leave the queue before submitting new jobs.
#syntax: "{default_queue_max_nb_jobs} queue1{max_nb_jobs1} queue2{max_nb_job2}"
OCFG_MAX_JOB_IN_QUEUE = 'MAX_JOB_IN_QUEUE' 

#database server
CFG_DATABASE_FILE = 'DATABASE_FILE'
CFG_TRANSFERED_FILES_DIR = 'TRANSFERED_FILES_DIR'
CFG_SERVER_NAME = 'SERVER_NAME'
CFG_NAME_SERVER_HOST ='NAME_SERVER_HOST'

OCFG_SERVER_LOG_FILE = 'SERVER_LOG_FILE'
OCFG_SERVER_LOG_LEVEL = 'SERVER_LOG_LEVEL'
OCFG_SERVER_LOG_FORMAT = 'SERVER_LOG_FORMAT'

#Engine
OCFG_ENGINE_LOG_DIR = 'ENGINE_LOG_DIR'
OCFG_ENGINE_LOG_LEVEL = 'ENGINE_LOG_LEVEL'
OCFG_ENGINE_LOG_FORMAT = 'ENGINE_LOG_FORMAT'

#Shared resource path translation files 
#specify the translation files (if any) associated to a namespace
#eg. translation_files = brainvisa{/home/toto/.brainvisa/translation.sjtr} namespace2{path/translation1.sjtr} namespace2{path/translation2.sjtr}
OCFG_PATH_TRANSLATION_FILES = 'PATH_TRANSLATION_FILES' 

# Soma-workflow light mode.
# Define this item to use soma-workflow in the light mode.
# This mode doesn't require a database server to run. It can not be used on
# remote computing resource. The client application can not be closed before the
# workflows and jobs are done. 
OCFG_LIGHT_MODE = 'LIGHT_MODE'

# Parallel job configuration :
# DRMAA attributes used in parallel job submission (their value depends on the cluster and DRMS) 
OCFG_PARALLEL_COMMAND = "drmaa_native_specification"
OCFG_PARALLEL_JOB_CATEGORY = "drmaa_job_category"
PARALLEL_DRMAA_ATTRIBUTES = [OCFG_PARALLEL_COMMAND, OCFG_PARALLEL_JOB_CATEGORY]
# kinds of parallel jobs (items can be added by administrator)
OCFG_PARALLEL_PC_MPI="MPI"
OCFG_PARALLEL_PC_OPEN_MP="OpenMP"
PARALLEL_CONFIGURATIONS = [OCFG_PARALLEL_PC_MPI, OCFG_PARALLEL_PC_OPEN_MP]
# parallel job environment variables for the execution machine (items can be added by administrators) 
OCFG_PARALLEL_ENV_MPI_BIN = 'PARALLEL_ENV_MPI_BIN'
OCFG_PARALLEL_ENV_NODE_FILE = 'PARALLEL_ENV_NODE_FILE'
PARALLEL_JOB_ENV = [OCFG_PARALLEL_ENV_MPI_BIN, OCFG_PARALLEL_ENV_NODE_FILE]

# Native_specification for all jobs
OCFG_NATIVE_SPECIFICATION = 'NATIVE_SPECIFICATION'

# local sheduler configuration -------------------------------------------------

OCFG_SCDL_CPU_NB = "CPU_NB"
OCFG_SCDL_INTERVAL = "SCHEDULER_INTERVAL"



#-------------------------------------------------------------------------------
# Classes and functions
#-------------------------------------------------------------------------------

class Configuration(observer.Observable):
  
  # path of the configuration file
  _config_path = None

  # config parser object
  _config_parser = None

  _resource_id = None

  _mode = None

  _scheduler_type = None

  _database_file = None

  _transfered_file_dir = None

  _submitting_machines = None

  _cluster_address = None

  _name_server_host = None

  _server_name = None

  _queue_limits = None

  _queues = None

  _drmaa_implementation = None

  _login = None

  _native_specification = None
  
  _sshport = 22
  
  _res_install_path = None

  parallel_job_config = None

  path_translation = None

  QUEUE_LIMITS_CHANGED = 0

  def __init__(self,
               resource_id,
               mode,
               scheduler_type,
               database_file,
               transfered_file_dir,
               submitting_machines=None,
               cluster_address=None,
               name_server_host=None,
               server_name=None,
               queues=None,
               queue_limits=None,
               drmaa_implementation=None,
               login=None,
               native_specification=None,
               sshport=22,
               res_install_path=None
               ):
    '''
    * resource_id *string*
      Identifier of the computing resource.

    * mode *string* 
      A mode among the existing modes defined in configuration.MODES
      
    * scheduler_type *string*
      A scheduler type among the existing schedulers defined in SCHEDULER_TYPES

    * database_file *string*
      Path of the database_file.

    * transfered_file_dir *string*
      Path of the directory where the transfered files are copied. Mandatory 
      in every mode (even local or light), the directory must exist.

    * submitting_machines *list of string*
      List of submitting machines. Mandatory in the REMOTE_MODE for the remote
      ssh connection.

    * cluster_address *string*
      Address of the cluster. Mandatory in the REMOTE_MODE for the remote
      ssh connection.

    * name_server_host *string*
      Machine where the pyro name server can be found. Mandatory in the 
      REMOTE_MODE to connect to the database_server.

    * server_name *string*
      Name of the database server regitered on the Pyro name server. Mandatory 
      in the REMOTE_MODE to connect to the database_server.
  
    * queues *list of string*
      List of the available queues. This item is only used in the GUI to make 
      easier the selection of the queue when submitting a workflow.

    * queue_limits *dictionary: string -> int*
      Maximum number of job in each queue (dictionary: queue name -> limit). 
      If a queue does not appear here, soma-workflow considers that there is 
      no limitation. 

    * drmaa_implementation *string*
      Set this item to "PBS" if you use FedStage PBS DRMAA 1.0 implementation, 
      otherwise it does not has to be set. 

    * login *string*
      Configure a login to avoid some typing in the GUI.

    * native_specification *string*
      Native specification applied to every jobs submitted to the resource 
      unless a different value is specified in the Job attribute 
      native_specification.
      
    * sshport *int*
      ssh port to remote machine.
    
    * res_install_path *string*
      soma-workflow installation path on the resource server
    '''

    super(Configuration, self).__init__()

    self._config_path = None
    self._config_parser = None
    self._resource_id = resource_id
    self._mode = mode
    self._scheduler_type = scheduler_type
    self._database_file = database_file
    self._transfered_file_dir = transfered_file_dir
    self._submitting_machines = submitting_machines
    self._cluster_address = cluster_address
    self._name_server_host = name_server_host
    self._server_name = server_name
    self._login = login
    self._native_specification = native_specification
    if queues == None:
      self._queues = []
    else:
      self._queues = queues
    if queue_limits == None:
      self._queue_limits = {}
    else:
      self._queue_limits = queue_limits
    self._queue_limits_disabled = False 
    self._drmaa_implementation = drmaa_implementation
    self.parallel_job_config = None
    self.path_translation = None
    
    self._sshport = sshport
    self._res_install_path =res_install_path


  @staticmethod
  def get_home_dir():
    # if python version > 2.5 expanduser should be ok
    homedir = os.getenv( 'HOME' )
    if not homedir:
      homedir = ''
      if sys.platform[:3] == 'win':
        homedir = os.getenv( 'USERPROFILE' )
        if not homedir:
          homedir = os.getenv( 'HOMEPATH' )
          if not homedir:
            homedir = '\\'
          drive = os.getenv( 'HOMEDRIVE' )
          if not drive:
            drive = os.getenv( 'SystemDrive' )
            if not drive:
              drive = os.getenv( 'SystemRoot' )
              if not drive:
                drive = os.getenv( 'windir' )
              if drive and len( drive ) >= 2:
                drive = drive[ :2 ]
              else:
                drive = ''
          homedir = drive + homedir
    return homedir


  @classmethod
  def load_from_file(cls,
                     resource_id=None, 
                     config_file_path=None):

    if config_file_path:
      config_path = config_file_path
    else:
      config_path = Configuration.search_config_path()

    config = None
    home_dir = Configuration.get_home_dir() 
    
    if resource_id == None or resource_id == socket.gethostname() :
      
      #scheduler local on the local machine
      resource_id = socket.gethostname()
      mode = LIGHT_MODE
      scheduler_type = LOCAL_SCHEDULER
      database_file = os.path.join(home_dir, ".soma-workflow/soma_workflow.db")
      transfered_file_dir = os.path.join(home_dir, ".soma-workflow/transfered_files")
      
      config = cls(resource_id=resource_id,
                   mode=mode,
                   scheduler_type=scheduler_type,
                   database_file=database_file,
                   transfered_file_dir=transfered_file_dir)

      if config_path != None:
        config_parser = ConfigParser.ConfigParser()
        config_parser.read(config_path)
        if config_parser.has_section(resource_id):
          config._config_parser = config_parser
          config._config_path = config_path

      if not os.path.isdir(os.path.join(home_dir, ".soma-workflow")):
        os.mkdir(os.path.join(home_dir, ".soma-workflow"))
      if not os.path.isdir(transfered_file_dir):
        os.mkdir(transfered_file_dir)

      return config

    else:

      if config_path == None:
        raise ConfigurationError("A configuration file is required to connect " 
                                 "to " + repr(resource_id) + ": the soma-workflow "
                                 "configuration file could not be found.")
      config_parser = ConfigParser.ConfigParser()
      config_parser.read(config_path)
      if not config_parser.has_section(resource_id):
        raise ConfigurationError("Can not find section " + repr(resource_id) + " "
                                "in configuration file: " + config_path)

      scheduler_type = None
      if config_parser.has_option(resource_id, OCFG_SCHEDULER_TYPE):
        scheduler_type = config_parser.get(resource_id, OCFG_SCHEDULER_TYPE)
        if scheduler_type not in SCHEDULER_TYPES:
          raise ConfigurationError("Unknown scheduler type:"
                                   " " + repr(scheduler_type) + 
                                   "Scheduler type must be one of the "
                                   "following:" + repr(SCHEDULER_TYPES))
      else:
        scheduler_type = DRMAA_SCHEDULER
        
      config = cls(resource_id=resource_id,
                   mode=None,
                   scheduler_type=scheduler_type,
                   database_file=None,
                   transfered_file_dir=None)
      config._config_parser = config_parser
      config._config_path = config_path

      return config


  
  def get_scheduler_type(self):
    return self._scheduler_type

  def get_res_install_path(self):
    if self._config_parser == None or self._res_install_path: 
      return self._res_install_path

    if not self._config_parser.has_option(self._resource_id,
                                          OCFG_INSTALLPATH):
      self._res_install_path = None
    else:
      self._res_install_path = self._config_parser.get(self._resource_id, OCFG_INSTALLPATH)
                        
    return self._res_install_path

  def get_ssh_port(self):
        if self._config_parser == None or self._sshport: 
          return self._sshport
    
        if self._config_parser.has_option(self._resource_id,
                                              OCFG_SSHPort):
          self._sshport = self._config_parser.get(self._resource_id, OCFG_SSHPort)
        else:
          self._sshport = "22"
                   
        return self._sshport


  def get_submitting_machines(self):
    if self._config_parser == None or self._submitting_machines: 
      return self._submitting_machines

    if not self._config_parser.has_option(self._resource_id,
                                          CFG_SUBMITTING_MACHINES):
      raise ConfigurationError("Can not find the configuration item %s for the "
                               "resource %s, in the configuration file %s." %
                               (CFG_SUBMITTING_MACHINES,
                                self._resource_id,
                                self._config_path))
    submitting_machines = self._config_parser.get(self._resource_id, 
                                         CFG_SUBMITTING_MACHINES).split()
    self._submitting_machines = []
    for sub_machine in submitting_machines:
      self._submitting_machines.append(os.path.expandvars(sub_machine))
    return self._submitting_machines


  @staticmethod
  def search_config_path():
    '''
    returns the path of the soma workflow configuration file
    '''

    home_dir = Configuration.get_home_dir() 

    config_path = os.getenv('SOMA_WORKFLOW_CONFIG')
    if not config_path or not os.path.isfile(config_path):
      config_path = os.path.join(home_dir, ".soma-workflow.cfg")
    if not config_path or not os.path.isfile(config_path):
      config_path = os.path.dirname(__file__)
      config_path = os.path.join(config_path, "etc/soma-workflow.cfg")
    if not config_path or not os.path.isfile(config_path):
      config_path = "/etc/soma-workflow.cfg"
    if not config_path or not os.path.isfile(config_path):
      config_path = None

    return config_path


  @staticmethod
  def get_configured_resources(config_file_path=None):
    '''
    returns the list of resouce ids
    '''
    resource_ids = []
    if config_file_path == None:
      return [socket.gethostname()]
    config_parser = ConfigParser.ConfigParser()
    config_parser.read(config_file_path)
    for r_id in config_parser.sections():
      resource_ids.append(r_id)
    local_resource_id = socket.gethostname()
    if not local_resource_id in resource_ids:
      resource_ids.append(local_resource_id)
    return resource_ids


  @staticmethod
  def get_logins(config_file_path=None):
    '''
    returns the dictionary resouce_id -> login
    '''
    resource_ids = []
    if config_file_path == None:
      return { socket.gethostname(): None }
    config_parser = ConfigParser.ConfigParser()
    config_parser.read(config_file_path)
    logins = {}
    for r_id in config_parser.sections():
      resource_ids.append(r_id)
      if config_parser.has_option(r_id, OCFG_LOGIN):
        logins[r_id] = config_parser.get(r_id, OCFG_LOGIN)
      else:
        logins[r_id] = None
    logins[socket.gethostname()] = None
    return logins

  def get_mode(self):
    '''
    Return the application mode: 'local', 'remote' or 'light'
    '''
    if self._mode:
      return self._mode

    if self._config_parser.has_option(self._resource_id, OCFG_LIGHT_MODE):
      self._mode = LIGHT_MODE
      return self._mode
    
    if not self._config_parser.has_option(self._resource_id, CFG_SUBMITTING_MACHINES):
       self._mode = LOCAL_MODE
       return self._mode
 
    if not self._submitting_machines:
      self._submitting_machines = self.get_submitting_machines()
   
    hostname = socket.gethostname()
    mode = REMOTE_MODE
    for machine in self._submitting_machines:
      if hostname == machine: 
        mode = LOCAL_MODE
    self._mode = mode 
    return mode
    

  def get_cluster_address(self):
    if self._config_parser == None or self._cluster_address:
      return self._cluster_address
    
    if not self._config_parser.has_option(self._resource_id, CFG_CLUSTER_ADDRESS):
      raise ConfigurationError("Can not find the configuration item %s for the "
                               "resource %s, in the configuration file %s." %
                               (CFG_CLUSTER_ADDRESS,
                                self._resource_id,
                                self._config_path))
    self._cluster_address= self._config_parser.get(self._resource_id, 
                                                   CFG_CLUSTER_ADDRESS)
    return self._cluster_address


  def get_database_file(self):
    if self._database_file:
      return self._database_file

    if not self._config_parser.has_option(self._resource_id, CFG_DATABASE_FILE):
      raise ConfigurationError("Can not find the configuration item %s " 
                                "for the resource %s, in the configuration " "file %s." %
                                (CFG_DATABASE_FILE,
                                  self._resource_id,
                                  self._config_path))
    self._database_file = self._config_parser.get(self._resource_id, 
                                                  CFG_DATABASE_FILE)
    self._database_file = os.path.expandvars(self._database_file)
    return self._database_file


  def get_transfered_file_dir(self):
    if self._transfered_file_dir:
      return self._transfered_file_dir

    if not self._config_parser.has_option(self._resource_id, 
                                          CFG_TRANSFERED_FILES_DIR):
      raise ConfigurationError("Can not find the configuration item %s " 
                                "for the resource %s, in the configuration " "file %s." %
                                (CFG_TRANSFERED_FILES_DIR,
                                  self._resource_id,
                                  self._config_path))
    self._transfered_file_dir = self._config_parser.get(self._resource_id, 
                                                      CFG_TRANSFERED_FILES_DIR)
    self._transfered_file_dir = os.path.expandvars(self._transfered_file_dir)
    return self._transfered_file_dir


  def get_parallel_job_config(self):
    if self._config_parser == None or self.parallel_job_config != None:
      return self.parallel_job_config
    
    self.parallel_job_config = {}
    for parallel_config_info in PARALLEL_DRMAA_ATTRIBUTES + \
                                PARALLEL_JOB_ENV + \
                                PARALLEL_CONFIGURATIONS:
      if self._config_parser.has_option(self._resource_id, parallel_config_info):
        self.parallel_job_config[parallel_config_info] = self._config_parser.get(self._resource_id, parallel_config_info)

    return self.parallel_job_config


  def get_drmaa_implementation(self):
    if self._config_parser == None or self._drmaa_implementation != None:
      return self._drmaa_implementation
    self._drmaa_implementation = None
    if self._config_parser != None and \
       self._config_parser.has_option(self._resource_id, 
                                      OCFG_DRMAA_IMPLEMENTATION):
      self._drmaa_implementation = self._config_parser.get(self._resource_id,                    
                                                      OCFG_DRMAA_IMPLEMENTATION)
    return self._drmaa_implementation

  def get_login(self):
    if self._config_parser == None or self._login != None:
      return self._login
    self._login = None
    if self._config_parser != None and \
       self._config_parser.has_option(self._resource_id, 
                                      OCFG_DRMAA_IMPLEMENTATION):
      self._login = self._config_parser.get(self._resource_id,                    
                                            OCFG_LOGIN)
    return self._login

  def get_native_specification(self):
    if self._config_parser == None or self._native_specification != None:
      return self._native_specification
    self._native_specification = None
    if self._config_parser != None and \
       self._config_parser.has_option(self._resource_id, 
                                      OCFG_NATIVE_SPECIFICATION):
      self._native_specification = self._config_parser.get(self._resource_id,                    
                                                    OCFG_NATIVE_SPECIFICATION)
    return self._native_specification




  def get_path_translation(self):
    if self._config_parser == None or self.path_translation != None:
      return self.path_translation

    self.path_translation = {}
    if self._config_parser.has_option(self._resource_id, 
                                      OCFG_PATH_TRANSLATION_FILES):
      translation_files_str = self._config_parser.get(self._resource_id, 
                                                    OCFG_PATH_TRANSLATION_FILES)
      #logger.info("Path translation files configured:")
      for ns_file_str in translation_files_str.split():
        ns_file = ns_file_str.split("{")
        namespace = ns_file[0]
        filename = ns_file[1].rstrip("}")
        filename = os.path.expandvars(filename)
        #logger.info(" -namespace: " + namespace + ", translation file: " + filename)
        try: 
          f = open(filename, "r")
        except IOError, e:
          raise ConfigurationError("Can not read the translation file %s" %
                                    (filename))
        
        if not namespace in self.path_translation.keys():
          self.path_translation[namespace] = {}
        line = f.readline()
        while line:
          splitted_line = line.split(None,1)
          if len(splitted_line) > 1:
            uuid = splitted_line[0]
            content = splitted_line[1].rstrip()
            #logger.info("    uuid: " + uuid + "   translation:" + content)
            self.path_translation[namespace][uuid] = os.path.expandvars(content)
          line = f.readline()
        f.close()

    return self.path_translation
      

  def get_name_server_host(self):
    if self._config_parser == None or self._name_server_host != None:
      return self._name_server_host
    self._name_server_host = None
    if self._config_parser != None:
      self._name_server_host = self._config_parser.get(self._resource_id, 
                                                       CFG_NAME_SERVER_HOST)
      self._name_server_host = os.path.expandvars(self._name_server_host)
    return self._name_server_host


  def get_server_name(self):
    if self._config_parser == None or self._server_name != None:
      return self._server_name
    if not self._config_parser.has_option(self._resource_id,
                                          CFG_SERVER_NAME):
      raise ConfigurationError("Can not find the configuration item %s " 
                                "for the resource %s, in the configuration " "file %s." %
                                (CFG_SERVER_NAME,
                                  self._resource_id,
                                  self._config_path))
    self._server_name = self._config_parser.get(self._resource_id, 
                                  CFG_SERVER_NAME)
    return self._server_name

  
  def change_queue_limits(self, queue_name, queue_limit):
    '''
    * queue_name *string*

    * queue_limit *int*
    '''
    self.get_queue_limits()
    self._queue_limits[queue_name] = queue_limit
    self.notifyObservers(Configuration.QUEUE_LIMITS_CHANGED)
 

  def disable_queue_limits(self):
    self._queue_limits_disabled = True 

  def get_queue_limits(self):
    if self._queue_limits_disabled:
      return {}

    if self._config_parser == None or len(self._queue_limits) != 0:
      return self._queue_limits

    self._queue_limits = {}
    if self._config_parser.has_option(self._resource_id, 
                                      OCFG_MAX_JOB_IN_QUEUE):
      queue_limits_str = self._config_parser.get(self._resource_id,
                                                 OCFG_MAX_JOB_IN_QUEUE)
      for info_str in queue_limits_str.split():
        info = info_str.split("{")
        if len(info[0]) == 0:
          queue_name = None
        else:
          queue_name = info[0]
        max_job = int(info[1].rstrip("}"))
        self._queue_limits[queue_name] = max_job

    return self._queue_limits

  
  def get_queues(self):
    if self._config_parser == None or len(self._queues) !=  0:
      return self._queues

    self._queues = []
    if self._config_parser.has_option(self._resource_id, OCFG_QUEUES):
      self._queues.extend(self._config_parser.get(self._resource_id,
                                                  OCFG_QUEUES).split())
    return self._queues


  def get_engine_log_info(self):
    if self._config_parser != None and self._config_parser.has_option(self._resource_id, OCFG_ENGINE_LOG_DIR):
      engine_log_dir = self._config_parser.get(self._resource_id, 
                                               OCFG_ENGINE_LOG_DIR)
      engine_log_dir = os.path.expandvars(engine_log_dir)
      if self._config_parser.has_option(self._resource_id,
                                        OCFG_ENGINE_LOG_FORMAT):
        engine_log_format = self._config_parser.get(self._resource_id,
                                                    OCFG_ENGINE_LOG_FORMAT,
                                                    1)
      else:
        engine_log_format = "%(asctime)s => %(module)s line %(lineno)s: %(message)s"
      if self._config_parser.has_option(self._resource_id,
                                        OCFG_ENGINE_LOG_LEVEL):
        engine_log_level = self._config_parser.get(self._resource_id,
                                                   OCFG_ENGINE_LOG_LEVEL)
      else:
        engine_log_level = "WARNING"
      return (engine_log_dir, engine_log_format, engine_log_level)
    else:
      return (None, None, None)



  def get_server_log_info(self):
    if self._config_parser != None and self._config_parser.has_option(self._resource_id, OCFG_SERVER_LOG_FILE):
      server_log_file = self._config_parser.get(self._resource_id,  
                                                OCFG_SERVER_LOG_FILE)
      server_log_file = os.path.expandvars(server_log_file)
      if self._config_parser.has_option(self._resource_id,
                                        OCFG_SERVER_LOG_FORMAT):
        server_log_format = self._config_parser.get(self._resource_id,
                                                    OCFG_SERVER_LOG_FORMAT, 
                                                    1)
      else:
        server_log_format = "%(asctime)s => %(module)s line %(lineno)s: %(message)s"
      if self._config_parser.has_option(self._resource_id,
                                        OCFG_SERVER_LOG_LEVEL):
        server_log_level = self._config_parser.get(self._resource_id,
                                                    OCFG_SERVER_LOG_LEVEL)
      else:
        server_log_level = "WARNING"
      return (server_log_file, server_log_format, server_log_level)
    else:
      return (None, None, None)


  def save_to_file(self, config_path=None):
    pass 

    #home_dir = Configuration.get_home_dir() 

    # disabled for now because it erases the commented part of the configuration
    #if not config_path:
      #if self._config_path != None:
        #config_path = self._config_path
      #else:
        #config_path = Configuration.search_config_path()
        #if config_path == None:
          #config_path = os.path.join(home_dir, ".soma-workflow.cfg")
        #print config_path

    #config_parser = ConfigParser.ConfigParser()
    #config_parser.read(config_path)

    #if not config_parser.has_section(self._resource_id):
      #config_file = open(config_path, "w")
      #config_parser.write(config_file)
      #config_file.close()
      #raise ConfigurationError("The configuration can not be saved."
                               #"The resource " + repr(self._resource_id) + " "
                               #"can not be found in the configuration "
                               #"file " + repr(config_path) + ".")

    #self.get_queue_limits()
    #if self._queue_limits != None and len(self._queue_limits):
      #queue_limits_str = ""
      #for queue, limit in self._queue_limits.iteritems():
        #if queue == None:
          #queue_limits_str = queue_limits_str + "{" + repr(limit) + "} "
        #else:
          #queue_limits_str = queue_limits_str + queue + "{" + repr(limit) + "} "
      #print "queue_limits_str " + queue_limits_str
      #config_parser.set(self._resource_id, 
                        #OCFG_MAX_JOB_IN_QUEUE, 
                        #queue_limits_str)

    #config_file = open(config_path, "w")
    #config_parser.write(config_file)
    #config_file.close() 


class LocalSchedulerCfg(observer.Observable):
  '''
  Local scheduler configuration.
  '''

  # number of processus which can run in parallel
  _proc_nb = None

  # interval (second)
  _interval = None

  # path of the configuration file
  _config_path = None


  PROC_NB_CHANGED = 0
  INTERVAL_CHANGED = 1

  def __init__(self, proc_nb=1, interval=1):
    '''
    * proc_nb *int*
      Number of processus which can run in parallel
    
    * interval *int*
      Update interval in second
    '''

    super(LocalSchedulerCfg, self).__init__()
    self._proc_nb = proc_nb
    self._interval = interval


  @classmethod
  def load_from_file(cls,
                     config_file_path=None):

    hostname = socket.gethostname()
    if config_file_path:
      config_path = config_file_path
    else:
      config_path = Configuration.search_config_path()

    config_parser = ConfigParser.ConfigParser()
    config_parser.read(config_path)

    if not config_parser.has_section(hostname):
      raise ConfigurationError("Wrong config file format. Can not find "
                               "section " + hostname + " in configuration " + "file: " + config_path)

    proc_nb = None
    interval = None
    
    if config_parser.has_option(hostname, 
                                OCFG_SCDL_CPU_NB):
      proc_nb_str = config_parser.get(socket.gethostname(),
                                      OCFG_SCDL_CPU_NB)
      proc_nb = int(proc_nb_str)
    if config_parser.has_option(hostname, 
                                OCFG_SCDL_INTERVAL):
      interval_str = config_parser.get(hostname,
                                       OCFG_SCDL_INTERVAL)
      interval = int(interval_str)

    if proc_nb == None and interval == None:
      config = cls()
    elif proc_nb == None and interval != None:
      config = cls(interval=interval)
    else:
      config = cls(proc_nb=proc_nb, interval=interval)
    config._config_path = config_path
    return config

  
  @staticmethod
  def search_config_path():
    '''
    returns the path of the soma workflow configuration file
    '''
    hostname = socket.gethostname()
    home_dir = Configuration.get_home_dir() 
    section_exist = False
    config_path = os.path.join(home_dir, ".soma-workflow-scheduler.cfg")
    if os.path.isfile(config_path):
      config_parser = ConfigParser.ConfigParser()
      config_parser.read(config_path)
      section_exist = config_parser.has_section(hostname)
    if not section_exist:
      config_path = os.path.dirname(__file__)
      config_path = os.path.join(config_path, "etc/soma-workflow-scheduler.cfg")
      if os.path.isfile(config_path):
        config_parser = ConfigParser.ConfigParser()
        config_parser.read(config_path)
        section_exist = config_parser.has_section(hostname)
    if not section_exist:
      config_path = "/etc/soma-workflow-scheduler.cfg"
      if os.path.isfile(config_path):
        config_parser = ConfigParser.ConfigParser()
        config_parser.read(config_path)
        section_exist = config_parser.has_section(hostname)
    if not section_exist:
      config_path = None

    return config_path

  def get_proc_nb(self):
    return self._proc_nb

  def get_interval(self):
    return self._interval

  def set_proc_nb(self, proc_nb):
    self._proc_nb = proc_nb
    self.notifyObservers(LocalSchedulerCfg.PROC_NB_CHANGED)
  
  def set_interval(self, interval):
    self._interval = interval
    self.notifyObservers(LocalSchedulerCfg.INTERVAL_CHANGED)

  def save_to_file(self, config_path=None):
    hostname = socket.gethostname()
    if not config_path:
      if self._config_path != None:
        
        config_path = self._config_path
      else:
        config_path = LocalSchedulerCfg.search_config_path()
        if config_path == None:
          home_dir = Configuration.get_home_dir()
          config_path = os.path.join(home_dir, ".soma-workflow-scheduler.cfg")

    config_parser = ConfigParser.ConfigParser()
    config_parser.read(config_path)

    if not config_parser.has_section(hostname):
      config_parser.add_section(hostname)

    config_parser.set(hostname, 
                      OCFG_SCDL_CPU_NB, 
                      str(self._proc_nb))
    config_parser.set(hostname, 
                      OCFG_SCDL_INTERVAL, 
                      str(self._interval))
    config_file = open(config_path, "w")
    config_parser.write(config_file)
    config_file.close()



def AddLineDefintions2BashrcFile(lines2add,path2bashrc=""):
    """Add line defintions to the ~/.bashrc file 
    
    Removing the lines2add which are already exsting in the ~/.bashrc
    Adding the lines2add at the end of ~/.bashrc
    
    Args:
        lines2add (string):  a list of line definitions. Line defintion like export PATH=~/mylocal/bin:${PATH}
        path2bashrc (string, optional): path to the ./bashrc file. It could be another path. Default paht is "~/.bashrc"

    Raises:
       IOError, ValueError
       
    Example:
        lines2add = ["SOMAWF_PATH=%s/soma-workflow"%(os.getenv("HOME")),
        'export PATH=$SOMAWF_PATH/bin:$PATH',
        'export PYTHONPATH=$SOMAWF_PATH/python:$PYTHONPATH',
        'export SOMA_WORKFLOW_EXAMPLES=$SOMAWF_PATH/test/jobExamples/',
        'export SOMA_WORKFLOW_EXAMPLES_OUT=$SOMAWF_PATH/test/jobExamples_out/']
        >>> print AddVariables2BashrcFile(lines2add, "~/.bashrc")
    """
    import os
    import sys
    
    if path2bashrc=="" :
        
        path2bashrc = os.path.join(os.getenv("HOME"),".bashrc")
        
    lines2rm = []
    content=[]

    try:
        with open(path2bashrc) as f:
            content = f.readlines()
    except IOError, e:
        print "I/O error({0}): {1}".format(e.errno, e.strerror)
        print "%s does not exist, the system will create the new file"% (path2bashrc)
    except ValueError:
        print "Could not convert data to an integer."
    except:
        print "Unexpected error:", sys.exc_info()[0]
        raise
    
    
    for i in range(len(content)):
        content[i]=content[i].strip()

    #try to find the duplicated paths and remove them
    for line2add in lines2add:
        isfound= any(line2add.strip()==cline.strip() for cline in content)
        while isfound:
            content.remove(line2add.strip())
            isfound= any(line2add.strip()==cline.strip() for cline in content)

    for line2add in lines2add:
        content.append(line2add)

    try:
        with open(path2bashrc,'w') as f:
            for cline in content:
                f.write(cline+"\n")
    except IOError, e:
        print "I/O error({0}): {1}".format(e.errno, e.strerror)
        print "The system cannot write the file %s. Please make sure it can be wrote. "% (path2bashrc)
        raise e
    except ValueError:
        print "Could not convert data to an integer."
        raise ValueError
    except:
        print "Unexpected error:", sys.exc_info()[0]
        raise

def WriteOutConfiguration(config_parser,config_path):
    try:
        with open(config_path,'w') as cfgfile:
            config_parser.write(cfgfile)
    except IOError, e:
        print "I/O error({0}): {1}".format(e.errno, e.strerror)
        print "The system cannot write the file %s. Please make sure that it can be written. "% (config_path)
        raise e
    except ValueError:
        print "Could not convert data to an integer.%s" % (config_path)
        raise ValueError
    except:
        print "Unexpected error:", sys.exc_info()[0]
        raise



from __future__ import with_statement

'''
@author: Benoit Da Mota
@author: Soizic Laguitton

@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}, 
@organization: PARIETAL, INRIA, Saclay, France

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''

import sys
import time
import threading
import logging
import os
import shutil
import optparse
import tarfile

from mpi4py import MPI

from soma_workflow import scheduler, constants


def slave_loop(communicator, 
               logger=None, 
               epd_to_deploy=None, 
               untar_directory='/tmp'):
    status = MPI.Status()
    rank = communicator.Get_rank()
    
    if not logger:
        logger = logging.getLogger("testMPI.slave")
    commands = {}

    if epd_to_deploy != None:
        lock_file_path = os.path.join(untar_directory, "sw_deploy_lock")
        if not os.path.isfile(lock_file_path):
            try:	
                lock_file = open(lock_file_path, "w")
                lock_file.write("locked \n")
                lock_file.close()
                epd_tar = tarfile.open(epd_to_deploy)
                epd_tar.extractall(path=untar_directory)
                #logger.debug('extract %s' %(epd_to_deploy))
            except IOError, e:
                logger.error("Could not deploy epd: %s" %(e))
                pass
        

    max_nb_jobs = 1

    ret_value = 0
    slave_on_hold = False
    while True:
        ended_jobs_info = {} # job_id -> (job_status, job_exit_status)
        t = None
        if len(commands) < max_nb_jobs:    
            communicator.send(dest=0,
                             tag=MPIScheduler.JOB_REQUEST)

            #logger.debug("Slave " + repr(rank) + " job request")
            communicator.Probe(source=MPI.ANY_SOURCE,
                               tag=MPI.ANY_TAG, status=status)
            t = status.Get_tag()
        elif communicator.Iprobe(source=MPI.ANY_SOURCE,
                               tag=MPI.ANY_TAG, status=status):
            t = status.Get_tag()
        if t != None:
            if t == MPIScheduler.JOB_SENDING:
                job_list = communicator.recv(source=0, tag=t)
                for j in job_list:
                    #process = scheduler.LocalScheduler.create_process(j)
                    separator = " "
                    command = ""
                    for command_el in j.plain_command():
                        command = command + "\'" + command_el + "\' "
                    #command = separator.join(j.plain_command())
                    logger.debug("Slave " + repr(rank) + " RUNS JOB" + repr(j.job_id) + " " + command)
                    commands[j.job_id] = command
            elif t == MPIScheduler.NO_JOB:
                communicator.recv(source=0, tag=t)
                #logger.debug("Slave " + repr(rank) + " "
                #             "received no job " + repr(commands))
                #time.sleep(1)
            elif t == MPIScheduler.EXIT_SIGNAL:
                communicator.send('STOP', dest=0, tag=MPIScheduler.EXIT_SIGNAL)
                logger.debug("Slave " + repr(rank) + " STOP !!!!! received")
                break
            elif t == MPIScheduler.JOB_KILL:
                job_ids = communicator.recv(source=0, tag=t)
                for job_id in job_ids:
                    if job_id in commands.keys():
                        # TO DO: relevant exception type and message
                        raise Exception("The job " + repr(job_id) + " can not be killed")

            else:
                raise Exception('Unknown tag')
        for job_id, command in commands.iteritems():
            if command == None:
                ended_jobs_info[job_id] = (constants.FAILED,
                                           (constants.EXIT_ABORTED, None, 
                                            None, None))
            else:

                if j.plain_stderr():
                    command = command + " >> " + j.plain_stdout() + " 2>> " + j.plain_stderr()
                else:
                    command = command +  " >> " + j.plain_stdout()

                ret_value = os.system(command)

                if ret_value != None:
                    #if ret_value != 0:		
                        #if ret_value == 32512:
                        #    slave_on_hold = True
                        #logger.debug("ERROR COMMAND %s, ret value: %s" %(command, repr(ret_value)))
                    ended_jobs_info[job_id] = (constants.DONE,
                                               (constants.FINISHED_REGULARLY, 
                                                ret_value, None, None))
     
        if ended_jobs_info:
            for job_id in ended_jobs_info.iterkeys():
                del commands[job_id]
            logger.debug("Slave " + repr(rank) + " send JOB_RESULT")
            communicator.send(ended_jobs_info, dest=0,
                              tag=MPIScheduler.JOB_RESULT)
        else:
            pass
            # TO DO send Slave is alive
        if slave_on_hold:
            logger.debug("Slave %d was punished !!!" %(rank))
            #time.sleep(1200) #20 mins 
            break
        else:
             time.sleep(1)


    if epd_to_deploy != None:
        logger.debug("Slave %d cleaning ... \n" %(rank))
        if os.path.isfile(lock_file_path):
           try:
                os.remove(lock_file_path)
                archive_dir_path = os.path.join(untar_directory, 'epd')
                if os.path.isdir(archive_dir_path):
                    shutil.rmtree(archive_dir_path)
                    logger.debug("remove %s" %(archive_dir_path))
           except Exception, e:
               pass
        logger.debug("Slave %d: end of cleaning! \n" %(rank))
    logger.debug("Slave %d END!!! \n" %(rank))

class MPIScheduler(scheduler.Scheduler):
    '''
    Allow to submit, kill and get the status of jobs.
    '''
    parallel_job_submission_info = None

    logger = None

    is_sleeping = None

    _proc_nb = None

    _queue = None

    _jobs = None

    _processes = None

    _status = None

    _exit_info = None

    _loop = None

    _interval = None

    _lock = None

    _failed_count = None

    JOB_REQUEST = 11
    JOB_SENDING = 12
    EXIT_SIGNAL = 13
    JOB_KILL = 14
    JOB_RESULT = 15
    NO_JOB = 16

    def __init__(self, communicator, interval=1, nb_attempt_per_job=1):
        super(MPIScheduler, self).__init__()

        self._communicator = communicator
        self.parallel_job_submission_info = None
        # self._proc_nb = proc_nb
        self._queue = []
        self._jobs = {}
        self._fail_count = {} # job_id -> nb of fail
        # self._processes = {}
        self._status = {}
        self._exit_info = {}
        self._lock = threading.RLock()
        self.stop_thread_loop = False
        self._interval = interval
 
        self._nb_attempt_per_job = nb_attempt_per_job

        self._logger = logging.getLogger("testMPI")
        def master_loop(self):
            self._stopped_slaves = 0
            while not self.stop_thread_loop:
                self._master_iteration()
                #time.sleep(0)#self._interval)

        self._loop = threading.Thread(name="scheduler_loop",
                                      target=master_loop,
                                      args=[self])
        self._loop.setDaemon(True)
        self._loop.start()


    def end_scheduler_thread(self):
        with self._lock:
            self.stop_thread_loop = True
            self._loop.join()
            print "Soma scheduler thread ended nicely."

    def _master_iteration(self):
        MPIStatus = MPI.Status()
        #if not self._queue:
        #    return
        self._communicator.Probe(source=MPI.ANY_SOURCE,
                                     tag=MPI.ANY_TAG,
                                     status=MPIStatus)
        with self._lock:
            t = MPIStatus.Get_tag()
            if t == MPIScheduler.JOB_REQUEST:
                #self._logger.debug("Master received the JOB_REQUEST signal")
                s = MPIStatus.Get_source()
                if not self._queue:
                    #self._logger.debug("Master No job for now")
                    self._communicator.recv(source=s, 
                                            tag=MPIScheduler.JOB_REQUEST)
                    self._communicator.send("No job for now", 
                                            dest=s,
                                            tag=MPIScheduler.NO_JOB)            
                else:
                    self._logger.debug("Master send a Job !!!")
                    self._communicator.recv(source=s, tag=MPIScheduler.JOB_REQUEST)
                    job_id = self._queue.pop(0)
                    job_list = [self._jobs[job_id]]
                    self._communicator.send(job_list, dest=s,
                                          tag=MPIScheduler.JOB_SENDING)
                    for j in job_list:
                        self._status[j.job_id] = constants.RUNNING
            elif t == MPIScheduler.JOB_RESULT:
                #self._logger.debug("Master received the JOB_RESULT signal")
                s = MPIStatus.Get_source()
                ended_jobs_info = self._communicator.recv(source=s,
                                                  tag=MPIScheduler.JOB_RESULT)
                for job_id, end_info in ended_jobs_info.iteritems():
                    job_status, exit_info = end_info
                    ret_value = exit_info[1]
                    if ret_value != 0 and \
                       (job_id not in self._fail_count or\
                       self._fail_count[job_id] < self._nb_attempt_per_job):
                        self._queue.insert(0, job_id)
                        if job_id in self._fail_count:
                            self._fail_count[job_id] = self._fail_count[job_id] + 1
                        else:
                            self._fail_count[job_id] = 1
                        self._logger.debug(repr(self._fail_count[job_id]) + " fails for job " + repr(job_id) + " ret value = " + repr(ret_value))
                    else:
                        self._exit_info[job_id] = exit_info
                        self._status[job_id] = job_status
            elif t == MPIScheduler.EXIT_SIGNAL:
                #self._logger.debug("Master received the EXIT_SIGNAL")
                self._stopped_slaves = self._stopped_slaves + 1
                if self._stopped_slaves == self._communicator.size -1:
                  self.stop_thread_loop = True
            else:
                self._logger.critical("Master unknown tag")

    def sleep(self):
        self.is_sleeping = True

    def wake(self):
        self.is_sleeping = False

    def clean(self):
        pass


    def queued_job_count(self):
        return len(self._queue)

    def job_submission(self, job):
        '''
        * job *EngineJob*
        * return: *string*
        Job id for the scheduling system (DRMAA for example)
        '''
        if not job.job_id or job.job_id == -1:
            raise Exception("Invalid job: no id")
        #self._logger.debug(">> job_submission wait lock")
        with self._lock:
            #self._logger.debug(">> job_submission wait lock END")
            self._queue.append(job.job_id)
            self._jobs[job.job_id] = job
            self._status[job.job_id] = constants.QUEUED_ACTIVE
            self._queue.sort(key=lambda job_id: self._jobs[job_id].priority,
                             reverse=True)
            self._logger.debug("A Job was submitted.")
        return job.job_id

    def get_job_status(self, scheduler_job_id):
        '''
        * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
        * return: *string*
        Job status as defined in constants.JOB_STATUS
        '''
        if not scheduler_job_id in self._status:
            raise Exception("Unknown job.")
        status = self._status[scheduler_job_id]
        return status

    def get_job_exit_info(self, scheduler_job_id):
        '''
        * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
        * return: *tuple*
        exit_status, exit_value, term_sig, resource_usage
        '''
        with self._lock:
            exit_info = self._exit_info[scheduler_job_id]
            del self._exit_info[scheduler_job_id]
        return exit_info

    def kill_job(self, scheduler_job_id):
        '''
        * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
        '''
        # TODO
        pass

if __name__ == '__main__':

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.size
        
    log_file_handler = logging.FileHandler(os.path.expandvars("$HOME/logtestmpi"))
    log_file_handler.setLevel(logging.DEBUG)
    log_formatter = logging.Formatter("%(asctime)s => %(module)s line %(lineno)s: %(message)s          %(threadName)s)")
    log_file_handler.setFormatter(log_formatter)

    parser = optparse.OptionParser()
    parser.add_option('--workflow',
                      dest='workflow_file', default=None,
                      help="The workflow to run.")
    parser.add_option('--restart', '-r', type="int",
                      dest='wf_id_to_restart', default=None,
                      help="The workflow id to restart")
    parser.add_option('--nb_attempt_per_job', type="int",
    	          dest='nb_attempt_per_job', default=1,
                      help="A job can be restarted several time if it fails. This option "
                           "specify the number of attempt per job. "
                           "By default, the jobs are not restarted.")
    group_alpha = optparse.OptionGroup(parser, "Alpha options")
    parser.add_option_group(group_alpha)
    group_alpha.add_option('--deploy_epd', dest='epd_to_deploy', default=None,
                      help="EPD tarball which will be inflated on each node using tar.")
    group_alpha.add_option('--untar_dir', dest='untar_directory', default='/tmp',
                      help="untar directory")
    
    options, args = parser.parse_args(sys.argv)

    if rank == 0:

        from soma_workflow.engine import WorkflowEngine, ConfiguredWorkflowEngine
        from soma_workflow.database_server import WorkflowDatabaseServer
        from soma_workflow.client import Helper
        import soma_workflow.configuration
               
        logger = logging.getLogger('testMPI')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(log_file_handler)

        resource_id = args[1]#sys.argv[1]
 
        if not len(args) == 2:
            #TO DO: stopping slave procedure in a function
            logger.critical("Mandatory argument: resource id.")
            for slave in range(1, comm.size):
                logger.debug("STOP !!!  slave " + repr(slave))
                comm.send('STOP', dest=slave, tag=MPIScheduler.EXIT_SIGNAL)
            logger.debug("######### master ends #############")
            raise Exception("Mandatory argument: resource id. \n")

        resource_id = args[1]
        
        try:
            config = soma_workflow.configuration.Configuration.load_from_file(resource_id)

            if config.get_scheduler_type() != soma_workflow.configuration.MPI_SCHEDULER:
                raise Exception("The resource id %s is not configured to use a MPI scheduler. "
                                "Configure a resource with SCHEDULER_TYPE = mpi." %(resource_id))
 
            logger.info(" ")
            logger.info(" ")
            logger.info(" ")
            logger.info(" ")
            logger.info(" ")
            logger.info(" ")
            logger.info("################ MASTER STARTS ####################")
 
            database_server = WorkflowDatabaseServer(config.get_database_file(),
                                                     config.get_transfered_file_dir())
   
            logger.info("workflow_file " + repr(options.workflow_file)) 
            logger.info("wf_id_to_restart " + repr(options.wf_id_to_restart)) 
            logger.info("nb_attempt_per_job " + repr(options.nb_attempt_per_job)) 
            logger.info("epd_to_deploy " + repr(options.epd_to_deploy)) 
            logger.info("untar_directory " + repr(options.untar_directory))
            sch = MPIScheduler(comm, interval=1, nb_attempt_per_job=options.nb_attempt_per_job)

            config.disable_queue_limits()    

            workflow_engine = ConfiguredWorkflowEngine(database_server,
                                                       sch,
                                                       config)
            if options.workflow_file and os.path.exists(options.workflow_file):
                workflow_file = options.workflow_file  
                logger.info(" ")
                logger.info("******* submission of worklfow **********")
                logger.info("workflow file: " + repr(workflow_file))

                workflow = Helper.unserialize(workflow_file)
                workflow_engine.submit_workflow(workflow,
                                                expiration_date=None,
                                                name=None,
                                                queue=None)
            if options.wf_id_to_restart != None:
                workflow_id = options.wf_id_to_restart
                logger.info(" ")
                logger.info("******* restart worklfow **********")
                logger.info("workflow if: " + repr(workflow_id))
                workflow_engine.stop_workflow(workflow_id)
                workflow_engine.restart_workflow(workflow_id, queue=None)
     
            while not workflow_engine.engine_loop.are_jobs_and_workflow_done():
                time.sleep(2)
            for slave in range(1, comm.size):
                logger.debug("STOP !!!  slave " + repr(slave))
                comm.send('STOP', dest=slave, tag=MPIScheduler.EXIT_SIGNAL)
            while not sch.stop_thread_loop:
                time.sleep(1)
            logger.debug("######### master ends #############")
        except Exception, e:
             for slave in range(1, comm.size):
                logger.debug("STOP !!!  slave " + repr(slave))
                comm.send('STOP', dest=slave, tag=MPIScheduler.EXIT_SIGNAL)
             raise e
    # slave code
    else:
        logger = logging.getLogger("testMPI.slave")
        logger.setLevel(logging.DEBUG)
        logger.addHandler(log_file_handler)
        logger.info("=====> slave starts " + repr(rank))
        slave_loop(comm, 
                   logger=logger, 
                   epd_to_deploy=options.epd_to_deploy,
                   untar_directory=options.untar_directory)



#! /usr/bin/env python

'''
@author: Jinpeng LI
@contact: mr.li.jinpeng@gmail.coms
@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''


from __future__ import with_statement

import os
import sys
from soma_workflow.connection import SSHExecCmd, check_if_somawfdb_on_server
import soma_workflow.configuration as configuration
from soma_workflow.configuration import WriteOutConfiguration
from soma_workflow.client import Job, Workflow, WorkflowController


def GetHostNameOnPBSTORQUE(userid, ip_address_or_domain, userpw=''):
    """To get a list of queue names on torque pbs

    Args:
       userid (str):  user name on the server side
       ip_address_or_domain (str): the ip address or the domain of the server
       userpw: user password to login the server using ssh. If you want to use "id_rsa.pub", just leave userpw = ""

    Returns:
       str.  hostname

    Raises:
       paramiko.AuthenticationException


    """
    sshcommand = "hostname"

    outlines = SSHExecCmd(sshcommand, userid, ip_address_or_domain, userpw)

    return outlines[0]


def GetQueueNamesOnPBSTORQUE(userid, ip_address_or_domain, userpw=''):
    """To get a list of queue names on torque pbs

    Args:
       userid (str):  user name on the server side
       ip_address_or_domain (str): the ip address or the domain of the server
       userpw: user password to login the server using ssh. If you want to use "id_rsa.pub", just leave userpw = ""

    Returns:
       list of str.  queue names

    Raises:
       paramiko.AuthenticationException


    """
    sshcommand = "qstat -Q"

    info_queue = []

    outlines = SSHExecCmd(sshcommand, userid, ip_address_or_domain, userpw)

    import re
    sline_idx = 0

    for sline in outlines:

        if sline_idx >= 2:  # skip the first line since it is the header
            sline = sline.strip()
            ssline = sline.split()
            if re.match("^[a-zA-Z]", ssline[0]):
                info_queue.append(ssline[0])

        sline_idx += 1

    # print repr(info_queue)
    return info_queue


def ConfiguratePaser(config_parser, resource_id, ip_address_or_domain, info_queue, userid, installpath, sshport):
    import soma_workflow.configuration as configuration

    oneline_queues = ''
    for one_q in info_queue:
        oneline_queues = oneline_queues+one_q+" "

    config_parser.add_section(resource_id)
    config_parser.set(
        resource_id, configuration.CFG_CLUSTER_ADDRESS, ip_address_or_domain)
    config_parser.set(
        resource_id, configuration.CFG_SUBMITTING_MACHINES, ip_address_or_domain)
    config_parser.set(resource_id, configuration.OCFG_QUEUES, oneline_queues)
    config_parser.set(resource_id, configuration.OCFG_LOGIN, userid)

    if sshport == None:
        sshport = 22
    config_parser.set(resource_id, configuration.OCFG_SSHPort, str(sshport))
    config_parser.set(resource_id, configuration.OCFG_INSTALLPATH, installpath)

    return config_parser


def RemoveResNameOnConfigureFile(resName, config_file_path=None):
    import sys
    import os
    import ConfigParser
    from ConfigParser import SafeConfigParser

    if config_file_path == None:
        config_file_path = configuration.Configuration.search_config_path()

    if config_file_path == None:
        return

    config_parser = SafeConfigParser()
    try:
        config_parser.read(config_file_path)
    except:
        strmsg = "Cannot open %s \n" % (config_file_path)
        sys.stderr.write(strmsg)
        raise Exception(strmsg)

    config_parser.remove_section(resName)

    WriteOutConfiguration(config_parser, config_file_path)

    return config_parser


def SetupConfigurationFileOnClient(resource_id,
                                   userid,
                                   ip_address_or_domain, userpw=None,
                                   install_swf_path_server=None,
                                   sshport=22):
    """To setup the configuration file on the client part

    Args:
       userid (str):  user name on the server side
       ip_address_or_domain (str): the ip address or the domain of the server
       userpw: user password to login the server using ssh. If you want to use "id_rsa.pub", just leave userpw = ""

    Raises:
       IOError, ValueError

    It will search the configuration file on the client.
    If it exists, it will keep the original configuration,
    if not, it will create the new configuration at $HOME/.soma-workflow.cfg

    """

    # ouput the configuration file
    import sys
    import os
    import ConfigParser
    from ConfigParser import SafeConfigParser

    config_file_path = configuration.Configuration.search_config_path()
    hostname = GetHostNameOnPBSTORQUE(userid, ip_address_or_domain, userpw)
    # resource_id="%s@%s"%(userid,hostname)
    # print "resource_id="+resource_id

    if config_file_path == None:

        home_dir = configuration.Configuration.get_home_dir()
        config_file_path = os.path.join(home_dir, ".soma-workflow.cfg")

        print "No configuration file on the client, we create a new one at %s" % (config_file_path)

        info_queue = GetQueueNamesOnPBSTORQUE(
            userid, ip_address_or_domain, userpw)
        config_parser = SafeConfigParser()

        config_parser = ConfiguratePaser(config_parser,
                                         resource_id,
                                         hostname,
                                         info_queue,
                                         userid,
                                         install_swf_path_server,
                                         sshport)
        WriteOutConfiguration(config_parser, config_file_path)

    else:
        print "Configuration file is found at: "+config_file_path
        config_parser = SafeConfigParser()
        config_parser.read(config_file_path)
        list_sections = config_parser.sections()

        if any(sec == resource_id for sec in list_sections):
            pass
        else:
            info_queue = GetQueueNamesOnPBSTORQUE(
                userid, ip_address_or_domain, userpw)
            config_parser = ConfiguratePaser(config_parser,
                                             resource_id,
                                             ip_address_or_domain,
                                             info_queue, userid,
                                             install_swf_path_server, sshport)
            WriteOutConfiguration(config_parser, config_file_path)


def CopySomaWF2Server(install_swf_path_server,
                      userid,
                      ip_address_or_domain,
                      userpw='', sshport=22):
    install_swf_path_server = os.path.join(install_swf_path_server, "python")
    install_swf_path_server_soma_workflow = os.path.join(
        install_swf_path_server, "./")
    install_swf_path_server_drmaa = os.path.join(install_swf_path_server)

    std_out_lines = SSHExecCmd(
        "mkdir -p '%s'" % (install_swf_path_server_drmaa),
        userid, ip_address_or_domain, userpw, sshport=sshport)
    print std_out_lines
    std_out_lines = SSHExecCmd(
        "mkdir -p '%s'" % (install_swf_path_server_soma_workflow),
        userid, ip_address_or_domain, userpw, sshport=sshport)
    print std_out_lines

    path2somawf = os.path.dirname(os.path.realpath(__file__))
    path2drmaa = os.path.join(path2somawf, "..", "somadrmaa")

    sshcommand = "scp -rC '%s' %s@%s:'%s'" % (path2somawf,
                                              userid,
                                              ip_address_or_domain,
                                              install_swf_path_server_soma_workflow)
    print "sshcommand="+sshcommand
    os.system(sshcommand)

    sshcommand = "scp -rC '%s' %s@%s:'%s'" % (path2drmaa,
                                              userid,
                                              ip_address_or_domain,
                                              install_swf_path_server_drmaa)
    print "sshcommand="+sshcommand
    os.system(sshcommand)

    SSHExecCmd("echo >> '%s'" % (os.path.join(
        install_swf_path_server_soma_workflow,
        "__init__.py")),
        userid,
        ip_address_or_domain,
        userpw,
        sshport=sshport)


def SimpleJobExample(resName, login, password):

    job_1 = Job(command=["sleep", "5"], name="job 1")
    job_2 = Job(command=["sleep", "5"], name="job 2")
    job_3 = Job(command=["sleep", "5"], name="job 3")
    job_4 = Job(command=["sleep", "5"], name="job 4")

    jobs = [job_1, job_2, job_3, job_4]
    dependencies = [(job_1, job_2),
                    (job_1, job_3),
                    (job_2, job_4),
                    (job_3, job_4)]

    workflow = Workflow(jobs=jobs,
                        dependencies=dependencies)

    controller = WorkflowController(resName, login, password)

    controller.submit_workflow(workflow=workflow,
                               name="TestConnectionExample")





def InstallSomaWF2Server(install_swf_path_server, ResName, userid, ip_address_or_domain, userpw='', sshport=22):

    CopySomaWF2Server(install_swf_path_server,
                      userid, ip_address_or_domain, userpw, sshport)

    script2install = os.path.join(
        install_swf_path_server, "python", "soma_workflow", "setup_server.py")

    command = "python '%s' -r %s " % (script2install, ResName)
    print "ssh command="+command

    SSHExecCmd(command, userid, ip_address_or_domain,
               userpw, wait_output=False, sshport=sshport)

    SetupConfigurationFileOnClient(
        ResName, userid, ip_address_or_domain, userpw, install_swf_path_server, sshport)

    os.system("sleep 5")

    # submit a workflow for test
    SimpleJobExample(ResName, userid, userpw)


def SetupSomaWF2Server(install_swf_path_server, ResName, userid, ip_address_or_domain, userpw='', sshport=22):

    if check_if_somawfdb_on_server(ResName, userid, ip_address_or_domain, userpw, sshport):
        msg = "Cannot find the soma-workflow on the server. Please first install soma-workflow"
        sys.stderr.write(msg)
        raise Exception(msg)

    SetupConfigurationFileOnClient(
        ResName, userid, ip_address_or_domain, userpw, install_swf_path_server, sshport)

    # submit a workflow for test
    SimpleJobExample(ResName, userid, userpw)


def RemoveSomaWF2Server(install_swf_path_server, ResName, userid, ip_address_or_domain, userpw='', sshport=22):

#    print "install_swf_path_server="+install_swf_path_server
#    print "ResName="+ResName
#    print "userid="+userid
#    print "ip_address_or_domain="+ip_address_or_domain
#    print "userpw="+userpw
#    print "sshport="+repr(sshport)

    # clean_script_server=os.path.join(install_swf_path_server,"python","soma","workflow","clean_server.py")

    clean_script_server = install_swf_path_server + \
        "/python/soma_workflow/clean_server.py"
    command = "python '%s' -r %s " % (clean_script_server, ResName)
    print command
    SSHExecCmd(command, userid, ip_address_or_domain,
               userpw, wait_output=False, sshport=sshport)
    os.system("sleep 5")

    command = "rm -rf '%s'" % (install_swf_path_server)
    print command
    SSHExecCmd(command, userid, ip_address_or_domain,
               userpw, wait_output=False, sshport=sshport)

    RemoveResNameOnConfigureFile(ResName)

    pass



# -----------------------------------------------------------
#  Copyright (C) 2009 StatPro Italia s.r.l. 
#                                                            
#  StatPro Italia                                            
#  Via G. B. Vico 4                                          
#  I-20123 Milano                                            
#  ITALY                                                     
#                                                            
#  phone: +39 02 96875 1                                     
#  fax:   +39 02 96875 605                                   
#                                                            
#  email: info@riskmap.net                                   
#                                                            
#  This program is distributed in the hope that it will be   
#  useful, but WITHOUT ANY WARRANTY; without even the        
#  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR   
#  PURPOSE. See the license for more details.                
# -----------------------------------------------------------
#  
#  Author: Enrico Sirola <enrico.sirola@statpro.com> 
#  $Id$ 

"""drmaa constants"""

# drmaa_get_attribute()
ATTR_BUFFER = 1024

# drmaa_get_contact() 
CONTACT_BUFFER = 1024

# drmaa_get_DRM_system() 
DRM_SYSTEM_BUFFER = 1024

# drmaa_get_DRM_system() 
DRMAA_IMPLEMENTATION_BUFFER = 1024

# Agreed buffer length constants 
# these are recommended minimum values 

ERROR_STRING_BUFFER =   1024
JOBNAME_BUFFER =        1024
SIGNAL_BUFFER =         32

# Agreed constants 
  
TIMEOUT_WAIT_FOREVER =  -1
TIMEOUT_NO_WAIT =       0 

JOB_IDS_SESSION_ANY =   "DRMAA_JOB_IDS_SESSION_ANY"
JOB_IDS_SESSION_ALL =   "DRMAA_JOB_IDS_SESSION_ALL"

SUBMISSION_STATE_ACTIVE = "drmaa_active"
SUBMISSION_STATE_HOLD =   "drmaa_hold"

# Agreed placeholder names
 
PLACEHOLDER_INCR =       "$drmaa_incr_ph$"
PLACEHOLDER_HD  =        "$drmaa_hd_ph$"
PLACEHOLDER_WD =         "$drmaa_wd_ph$"

# Agreed names of job template attributes 

REMOTE_COMMAND =        "drmaa_remote_command"
JS_STATE =              "drmaa_js_state"
WD =                    "drmaa_wd"
JOB_CATEGORY =          "drmaa_job_category"
NATIVE_SPECIFICATION =  "drmaa_native_specification"
BLOCK_EMAIL =           "drmaa_block_email"
START_TIME =            "drmaa_start_time"
JOB_NAME =              "drmaa_job_name"
INPUT_PATH =            "drmaa_input_path"
OUTPUT_PATH =           "drmaa_output_path"
ERROR_PATH =            "drmaa_error_path"
JOIN_FILES =            "drmaa_join_files"
TRANSFER_FILES =        "drmaa_transfer_files"
DEADLINE_TIME =         "drmaa_deadline_time"
WCT_HLIMIT =            "drmaa_wct_hlimit"
WCT_SLIMIT =            "drmaa_wct_slimit"
DURATION_HLIMIT =       "drmaa_duration_hlimit"
DURATION_SLIMIT =       "drmaa_duration_slimit"

# names of job template vector attributes 
V_ARGV =                "drmaa_v_argv"
V_ENV =                 "drmaa_v_env"
V_EMAIL =               "drmaa_v_email"

NO_MORE_ELEMENTS = 25

def job_state(code):
    return _JOB_PS[code]

class JobState(object):
    UNDETERMINED='undetermined'
    QUEUED_ACTIVE='queued_active'
    SYSTEM_ON_HOLD='system_on_hold'
    USER_ON_HOLD='user_on_hold'
    USER_SYSTEM_ON_HOLD='user_system_on_hold'
    RUNNING='running'
    SYSTEM_SUSPENDED='system_suspended'
    USER_SUSPENDED='user_suspended'
    USER_SYSTEM_SUSPENDED='user_system_suspended'
    DONE='done'
    FAILED='failed'

# Job control action
class JobControlAction(object):
    SUSPEND='suspend'
    RESUME='resume'
    HOLD='hold'
    RELEASE='release'
    TERMINATE='terminate'  

_JOB_CONTROL = [
    JobControlAction.SUSPEND,
    JobControlAction.RESUME,  
    JobControlAction.HOLD,    
    JobControlAction.RELEASE, 
    JobControlAction.TERMINATE
]

def string_to_control_action(operation):
    return _JOB_CONTROL.index(operation)
def control_action_to_string(code):
    return _JOB_CONTROL[code]
def status_to_string(status):
    return _JOB_PS[status]

_JOB_PS = { 
    0x00: JobState.UNDETERMINED,
    0x10: JobState.QUEUED_ACTIVE        ,
    0x11: JobState.SYSTEM_ON_HOLD       ,
    0x12: JobState.USER_ON_HOLD         ,
    0x13: JobState.USER_SYSTEM_ON_HOLD  ,
    0x20: JobState.RUNNING                                   ,
    0x21: JobState.SYSTEM_SUSPENDED     ,
    0x22: JobState.USER_SUSPENDED       ,
    0x23: JobState.USER_SYSTEM_SUSPENDED,
    0x30: JobState.DONE                 ,
    0x40: JobState.FAILED               ,
}                  

# State at submission time
class JobSubmissionState(object):
    HOLD_STATE=SUBMISSION_STATE_HOLD
    ACTIVE_STATE=SUBMISSION_STATE_ACTIVE

_SUBMISSION_STATE = [
    JobSubmissionState.HOLD_STATE,
    JobSubmissionState.ACTIVE_STATE
]

def submission_state(code):
    return _SUBMISSION_STATE[code]




'''
@author: Soizic Laguitton

@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''


'''
Engine jobs, workflows and file transfers are defined in a separated 
module because of the dependencies of the engine module.
The class EngineJob, EngineWorkflow and EngineTransfer can thus be 
uses on the client side (in the GUI for example) without importing
module required in the engine module. 
'''

import types
import os
import logging

from soma_workflow.errors import JobError, WorkflowError
import soma_workflow.constants as constants
from soma_workflow.client import Job, FileTransfer, Workflow, SharedResourcePath, Group


class EngineJob(Job):
  
  # job id
  job_id = None
  # workflow id 
  workflow_id = None
  # user_id
  _user_id = None  
  # string
  drmaa_id = None
  # name of the queue to be used to submit jobs, str
  queue = None
  # job status as defined in constants.JOB_STATUS. string
  status = None
  # last status update date
  last_status_update = None
  # exit status string as defined in constants. JOB_EXIT_STATUS
  exit_status = None
  # contains operating system exit value if the status is FINISHED_REGULARLY. 
  # int or None
  exit_value = None

  str_rusage = None
  # contain a representation of the signal if the status is FINISHED_TERM_SIG.
  # string or None
  terminating_signal = None

  expiration_date = None

  # mapping between FileTransfer and actual EngineTransfer which are valid on
  # the system. 
  # dictionary: FileTransfer -> EngineTransfer
  transfer_mapping = None

  # mapping between SharedResourcePath and actual path which are valid on the
  # system
  # dictonary: SharedResourcePath -> string (path)
  srp_mapping = None

  path_translation = None

  logger = None

  def __init__(self, 
               client_job, 
               queue, 
               workflow_id=-1, 
               path_translation=None,
               transfer_mapping=None,
               parallel_job_submission_info=None):

    super(EngineJob, self).__init__(client_job.command,
                                    client_job.referenced_input_files,
                                    client_job.referenced_output_files ,
                                    client_job.stdin,
                                    client_job.join_stderrout,
                                    client_job.disposal_timeout,
                                    client_job.name,
                                    client_job.stdout_file,
                                    client_job.stderr_file ,
                                    client_job.working_directory ,
                                    client_job.parallel_job_info,
                                    client_job.priority,
                                    client_job.native_specification)
    
    self.job_id = -1

    self.drmaa_id = None
    self.status = constants.NOT_SUBMITTED
    self.exit_status = None
    self.exit_value = None
    self.terminating_signal = None

    self.workflow_id = workflow_id
    self.queue = queue

    self.path_translation = path_translation

    if not transfer_mapping:
      self.transfer_mapping = {}
    else:
      self.transfer_mapping = transfer_mapping
    self.srp_mapping = {}

    self._map(parallel_job_submission_info)

    
  def _map(self, parallel_job_submission_info):
    '''
    Fill the transfer_mapping and srp_mapping attributes.
    + check the types of the Job arguments.
    '''
    if not self.command:
      raise JobError("The command attribute is the only required "
                     "attribute of Job.")

    if self.parallel_job_info:
      parallel_config_name, max_node_number = self.parallel_job_info
      if not parallel_job_submission_info:
        raise JobError("No parallel information was registered for the "
                       " current resource. A parallel job can not be submitted")
      if parallel_config_name not in parallel_job_submission_info:
        raise JobError("The parallel job can not be submitted because the "
                        "parallel configuration %s is missing." %(configuration_name))


    if self.stdin:
      if isinstance(self.stdin, FileTransfer):
        if not self.stdin in self.referenced_input_files:
          self.referenced_input_files.append(self.stdin)
        #if not self.stdin in self.transfer_mapping:
          #if isinstance(self.stdin, EngineTransfer):
            ## TBI check that the transfer exist in the database 
            #self.transfer_mapping[self.stdin] = self.stdin
          #else:
            #eft = EngineTransfer(self.stdin)
            #self.transfer_mapping[self.stdin] = eft
      elif isinstance(self.stdin, SharedResourcePath):
        self.srp_mapping[self.stdin] = self._translate(self.stdin) 
      else:
        if not type(self.stdin) in types.StringTypes:
          raise JobError("Wrong stdin type: %s" %(repr(self.stdin))) 
        self.stdin = os.path.abspath(self.stdin)

    if self.working_directory:
      if isinstance(self.working_directory, FileTransfer):
        if not self.working_directory in self.referenced_input_files:
          self.referenced_input_files.append(self.working_directory)
        if not self.working_directory in self.referenced_output_files:
          self.referenced_output_files.append(self.working_directory)
        #if not self.working_directory in self.transfer_mapping:
          #if isinstance(self.working_directory, EngineTransfer):
            ## TBI check that the transfer exist in the database 
            #self.transfer_mapping[self.working_directory] = self.working_directory
          #else:
            #eft = EngineTransfer(self.working_directory)
            #self.transfer_mapping[self.working_directory] = eft
      elif isinstance(self.working_directory, SharedResourcePath):
        self.srp_mapping[self.working_directory] = self._translate(self.working_directory)
      else:
        if not type(self.working_directory) in types.StringTypes:
          raise JobError("Wrong working directory type: %s " %
                         (repr(self.working_directory)))
        self.working_directory = os.path.abspath(self.working_directory)

    if self.stdout_file:
      if isinstance(self.stdout_file, FileTransfer):
        if not self.stdout_file in self.referenced_output_files:
          self.referenced_output_files.append(self.stdout_file)
        #if not self.stdout_file in self.transfer_mapping:
          #if isinstance(self.stdout_file, EngineTransfer):
            ## TBI check that the transfer exist in the database 
            #self.transfer_mapping[self.stdout_file] = self.stdout_file
          #else:
            #eft = EngineTransfer(self.stdout_file)
            #self.transfer_mapping[self.stdout_file] = eft
      elif isinstance(self.stdout_file, SharedResourcePath):
        self.srp_mapping[self.stdout_file] = self._translate(self.stdout_file) 
      else:
        if not type(self.stdout_file) in types.StringTypes:
          raise JobError("Wrong stdout_file type: %s" %(repr(self.stdout_file))) 
        self.stdout_file = os.path.abspath(self.stdout_file)

    if self.stderr_file:
      if isinstance(self.stderr_file, FileTransfer):
        if not self.stderr_file in self.referenced_output_files:
          self.referenced_output_files.append(self.stderr_file)
        #if not self.stderr_file in self.transfer_mapping:
          #if isinstance(self.stderr_file, EngineTransfer):
            ## TBI check that the transfer exist in the database 
            #self.transfer_mapping[self.stderr_file] = self.stderr_file
          #else:
            #eft = EngineTransfer(self.stderr_file)
            #self.transfer_mapping[self.stderr_file] = eft
      elif isinstance(self.stderr_file, SharedResourcePath):
        self.srp_mapping[self.stderr_file] = self._translate(self.stderr_file) 
      else:
        if not type(self.stderr_file) in types.StringTypes:
          raise JobError("Wrong stderr_file type: %s" %(repr(self.stderr_file))) 
        self.stderr_file = os.path.abspath(self.stderr_file)
    

    # transfer_mapping from referenced_input_files and referenced_output_files
    # + type checking
    for ft in self.referenced_input_files:
      if not isinstance(ft, FileTransfer):
        raise JobError("%s: Wrong type in referenced_input_files. "
                       " FileTransfer object required." %(repr(ft)))
      elif isinstance(ft, EngineTransfer) and \
          ft not in self.transfer_mapping:
        # TBI check that the transfer exist in the database 
        self.transfer_mapping[ft] = ft
      elif ft not in self.transfer_mapping:
        eft = EngineTransfer(ft)
        self.transfer_mapping[ft] = eft
   
    for ft in self.referenced_output_files:
      if not isinstance(ft, FileTransfer):
        raise JobError("%s: Wrong type in referenced_output_files. "
                       " FileTransfer object required." %(repr(ft)))
      elif isinstance(ft, EngineTransfer) and \
          ft not in self.transfer_mapping:
        # TBI check that the transfer exist in the database 
        self.transfer_mapping[ft] = ft
      elif ft not in self.transfer_mapping:
        eft = EngineTransfer(ft)
        self.transfer_mapping[ft] = eft

    # filling the srp_mapping exploring the command, stdin, stdout_file, 
    #stderr_file, workflow_directory 
    # + type checking
    for command_el in self.command:
      if isinstance(command_el, SharedResourcePath):
        self.srp_mapping[command_el] = self._translate(command_el)
      elif isinstance(command_el, FileTransfer):
        if not command_el in self.transfer_mapping:
          raise JobError("The FileTransfer objets used in the "
                         "command must be declared in the Job "
                         "attributes: referenced_input_files "
                         "and referenced_output_files.")

      elif isinstance(command_el, tuple) and \
           isinstance(command_el[0], FileTransfer):
        if not command_el[0] in self.transfer_mapping:
          raise JobError("The FileTransfer objets used in the "
                         "command must be declared in the Job "
                         "attributes: referenced_input_files "
                         "and referenced_output_files.")
      elif isinstance(command_el, list):
        new_list = []
        for list_el in command_el:
          if isinstance(list_el, SharedResourcePath):
            self.srp_mapping[list_el] = self._translate(list_el) 
          elif isinstance(list_el, FileTransfer):
            if not list_el in self.transfer_mapping:
              raise JobError("The FileTransfer objets used in the "
                             "command must be declared in the Job "
                             "attributes: referenced_input_files "
                             "and referenced_output_files.")
          elif isinstance(list_el, tuple) and \
              isinstance(list_el[0], FileTransfer):
            if not list_el[0] in self.transfer_mapping:
              raise JobError("The FileTransfer objets used in the "
                             "command must be declared in the Job "
                             "attributes: referenced_input_files "
                             "and referenced_output_files.")
          else:
            if not type(list_el) in types.StringTypes:
              raise JobError("Wrong command element type: %s" %(repr(list_el)))
      else:
        if not type(command_el) in types.StringTypes:
          raise JobError("Wrong command element type: %s" %(repr(command_el)))
      
 

  def _translate(self, srp):
    '''
    srp: SharedResourcePath
    returns: the translated path
    '''
    
    if not self.path_translation:
      raise JobError("Could not submit workflows or jobs with shared resource "
                     "path: no translation found in the configuration file.")
    if not srp.namespace in self.path_translation.keys():
      raise JobError("Could not translate shared resource path. The "               
                     "namespace %s is not configured." %(srp.namespace))
    if not srp.uuid in self.path_translation[srp.namespace]:
      raise JobError("Could not translate shared resource path. The uuid %s "
                     "does not exist for the namespace %s." %
                     (srp.uuid, srp.namespace))
    translated_path = os.path.join(self.path_translation[srp.namespace][srp.uuid],
                                   srp.relative_path)
    return translated_path

  def plain_command(self):
    '''
    Compute the actual job command (sequence of string) from the command 
    holding FileTransfer and SharedResourcePath objects.

    returns: sequence of string
    '''
    plain_command = []
    for command_el in self.command:
      if isinstance(command_el, SharedResourcePath):
        plain_command.append(self.srp_mapping[command_el])
      elif isinstance(command_el, FileTransfer):
        plain_command_el = self.transfer_mapping[command_el].engine_path
        plain_command.append(plain_command_el)
      elif isinstance(command_el, tuple) and \
           isinstance(command_el[0], FileTransfer):
        plain_command_el = self.transfer_mapping[command_el[0]].engine_path
        plain_command_el = os.path.join(plain_command_el, command_el[1])
        plain_command.append(plain_command_el)
      elif isinstance(command_el, list):
        new_list = []
        for list_el in command_el:
          if isinstance(list_el, SharedResourcePath):
            new_list.append(self.srp_mapping[list_el])
          elif isinstance(list_el, FileTransfer):
            new_list_el = self.transfer_mapping[list_el].engine_path
            new_list.append(new_list_el)
          elif isinstance(list_el, tuple) and \
              isinstance(list_el[0], FileTransfer):
            new_list_el = self.transfer_mapping[list_el[0]].engine_path
            new_list_el = os.path.join(new_list_el, list_el[1])
            new_list.append(new_list_el)
          else:
            assert(type(list_el) in types.StringTypes)
            new_list.append(list_el)
        str_list = str(repr(new_list)) 
        plain_command.append(str_list.replace("'", "\""))
      else:
        assert(type(command_el) in types.StringTypes)
        plain_command.append(command_el)
    return plain_command
    
  def plain_stdin(self):
    if self.stdin and isinstance(self.stdin, FileTransfer):
      return self.transfer_mapping[self.stdin].engine_path
    if self.stdin and isinstance(self.stdin, SharedResourcePath):
      return self.srp_mapping[self.stdin]
    return self.stdin

  def plain_stdout(self):
    if self.stdout_file and isinstance(self.stdout_file, FileTransfer):
      return self.transfer_mapping[self.stdout_file].engine_path
    if self.stdout_file and isinstance(self.stdout_file, SharedResourcePath):
      return self.srp_mapping[self.stdout_file]
    return self.stdout_file

  def plain_stderr(self):
    if self.stderr_file and isinstance(self.stderr_file, FileTransfer):
      return self.transfer_mapping[self.stderr_file].engine_path
    if self.stderr_file and isinstance(self.stderr_file, SharedResourcePath):
      return self.srp_mapping[self.stderr_file]
    return self.stderr_file

  def plain_working_directory(self):
    if self.working_directory and \
       isinstance(self.working_directory, FileTransfer):
      return self.transfer_mapping[self.working_directory].engine_path
    if self.working_directory and \
       isinstance(self.working_directory, SharedResourcePath):
      return self.srp_mapping[self.working_directory]
    return self.working_directory


  def is_running(self):
    running = self.status != constants.NOT_SUBMITTED and \
              self.status != constants.FAILED and \
              self.status != constants.DONE 
    return running

  def is_done(self):
    done = self.status == constants.DONE or self.status == constants.FAILED
    return done

  def failed(self):
    failed = (self.is_done() and \
              ((self.exit_value != 0 and self.exit_value != None) or \
              self.exit_status != constants.FINISHED_REGULARLY or \
              self.terminating_signal != None)) 
    return failed

  def ended_with_success(self):
    success = self.is_done() and \
              self.exit_value == 0 and \
              self.exit_status == constants.FINISHED_REGULARLY and \
              self.terminating_signal == None
    return success



class EngineWorkflow(Workflow):
  
  # workflow id
  wf_id = None
  # user id
  _user_id = None
  # path translation for each namespace a dictionary holding the traduction 
  #(association uuid => engine path)
  # dictionary, namespace => uuid => path
  _path_translation = None
  # workflow status as defined in constants.WORKFLOW_STATUS
  status = None
  # expidation date
  expiration_date = None
  # name of the queue to be used to submit jobs, str
  queue = None
  
  # mapping between Job and actual EngineJob which are valid on the system
  # dictionary: Job -> EngineJob
  job_mapping = None
  
  # mapping between FileTransfer and actual EngineTransfer which are valid on
  # the system. 
  # dictionary: FileTransfer -> EngineTransfer
  transfer_mapping = None

  # Once registered on the database server each 
  # EngineJob has an job_id.
  # dictonary: job_id -> EngineJob
  registered_jobs = None
  
  # Once registered on the database server each 
  # EngineTransfer has an transfer_id.
  # dictonary: tr_id -> EngineTransfer
  registered_tr = None

  # for each job: list of all the jobs which have to end before a job can start
  #dictionary: job_id -> list of job id
  _dependency_dict = None

  logger = None
  
  def __init__(self, 
               client_workflow, 
               path_translation, 
               queue, 
               expiration_date, 
               name):
 
    super(EngineWorkflow, self).__init__(client_workflow.jobs,
                                         client_workflow.dependencies,
                                         client_workflow.root_group,
                                         client_workflow.groups)
    self.wf_id = -1

    self.status = constants.WORKFLOW_NOT_STARTED
    self._path_translation = path_translation
    self.queue = queue
    self.expiration_date = expiration_date
    self.name = name

    self.user_storage = client_workflow.user_storage

    self.job_mapping = {}
    self.transfer_mapping = {}
    self._map()

    self.registered_tr = {}
    self.registered_jobs = {}

    self._dependency_dict = {}
    for dep in self.dependencies:
      if dep[1] in self._dependency_dict:
        self._dependency_dict[dep[1]].append(dep[0])
      else:
        self._dependency_dict[dep[1]] = [dep[0]] 




  def _map(self):
    '''
    Fill the job_mapping attributes.
    + type checking
    '''
    # jobs
    for job in self.jobs:
      if not isinstance(job, Job):
        raise WorkflowError("%s: Wrong type in the jobs attribute. "
                            " An object of type Job is required." %(repr(job)))
      if job not in self.job_mapping:
        ejob = EngineJob(client_job=job,
                         queue=self.queue,
                         path_translation=self._path_translation,
                         transfer_mapping=self.transfer_mapping)
        self.transfer_mapping.update(ejob.transfer_mapping)
        self.job_mapping[job]=ejob

    # dependencies
    for dependency in self.dependencies:
      if not isinstance(dependency[0], Job) or \
         not isinstance(dependency[1], Job):
          raise WorkflowError("%s, %s: Wrong type in the workflow dependencies."
                              " An object of type Job is required." %
                              (repr(dependency[0]), repr(dependency[1])))


      if dependency[0] not in self.job_mapping:
        self.jobs.append(dependency[0])
        ejob = EngineJob( client_job=dependency[0],
                          queue=self.queue,
                          path_translation=self._path_translation,
                          transfer_mapping=self.transfer_mapping)
        self.transfer_mapping.update(ejob.transfer_mapping)
        self.job_mapping[dependency[0]]=ejob

      if dependency[1] not in self.job_mapping:
        self.jobs.append(dependency[1])
        ejob = EngineJob( client_job=dependency[1],
                          queue=self.queue,
                          path_translation=self._path_translation,
                          transfer_mapping=self.transfer_mapping)
        self.transfer_mapping.update(ejob.transfer_mapping)
        self.job_mapping[dependency[1]]=ejob
      
    # groups
    groups = self.groups 
    for group in self.groups:
      for elem in group.elements:
        if isinstance(elem, Job):
          if elem not in self.job_mapping:
            self.jobs.append(elem)
            ejob = EngineJob( client_job=elem,
                              queue=self.queue,
                              path_translation=self._path_translation,
                              transfer_mapping=self.transfer_mapping)
            self.transfer_mapping.update(ejob.transfer_mapping)
            self.job_mapping[elem]=ejob
        elif not isinstance(elem, Group):
          raise WorkflowError("%s: Wrong type in the workflow "
                              "groups. Objects of type Job or " 
                              "Group are required." %(repr(elem)))

    # root group
    for elem in self.root_group:
      if isinstance(elem, Job):
        if elem not in self.job_mapping:
          self.jobs.append(elem)
          ejob = EngineJob( client_job=elem,
                            queue=self.queue,
                            path_translation=self._path_translation,
                            transfer_mapping=self.transfer_mapping)
          self.transfer_mapping.update(ejob.transfer_mapping)
          self.job_mapping[elem]=ejob
      elif not isinstance(elem, Group):
        raise WorkflowError("%s: Wrong type in the workflow root_group."
                            " Objects of type Job or Group are required." %
                            (repr(elem)))
    

  def find_out_independant_jobs(self):
    independant_jobs = []
    for job in self.jobs:
      to_run=True
      for ft in job.referenced_input_files:
        if not self.transfer_mapping[ft].files_exist_on_server():
          if self.transfer_mapping[ft].status == constants.TRANSFERING_FROM_CR_TO_CLIENT:
              #TBI stop the transfer
              pass 
          to_run = False
          break 
      if to_run:
        for dep in self.dependencies:
          if dep[1] == job:
            to_run = False
            break
      if to_run:
        independant_jobs.append(self.job_mapping[job])
    if independant_jobs:
      status = constants.WORKFLOW_IN_PROGRESS
    else:
      status = self.status
    return (independant_jobs, status)

  def find_out_jobs_to_process(self):
    '''
    Workflow exploration to find out new node to process.

    @rtype: tuple (sequence of EngineJob,
                   sequence of EngineJob,
                   constanst.WORKFLOW_STATUS)
    @return: (jobs to run,
              ended jobs
              workflow status)
    '''

    self.logger = logging.getLogger('engine.EngineWorkflow') 
    self.logger.debug("self.jobs="+repr(self.jobs))
    to_run = []
    to_abort = set([])
    done = []
    running = []
    for client_job in self.jobs:
      self.logger.debug("client_job="+repr(client_job))
      job = self.job_mapping[client_job]
      if job.is_done(): 
        done.append(job)
      elif job.is_running(): 
        running.append(job)
      elif job.status == constants.NOT_SUBMITTED:
        job_to_run = True
        job_to_abort = False
        for ft in job.referenced_input_files:
          eft = job.transfer_mapping[ft]
          if not eft.files_exist_on_server():
            if eft.status == constants.TRANSFERING_FROM_CR_TO_CLIENT:
              #TBI stop the transfer
              pass 
            job_to_run = False
            break
        if client_job in self._dependency_dict:
          for dep_client_job in self._dependency_dict[client_job]:
            dep_job = self.job_mapping[dep_client_job]
            if not dep_job.ended_with_success():
              job_to_run = False
              if dep_job.failed():
                job_to_abort = True
                break
            # TO DO to abort
        if job_to_run:
          to_run.append(job)
        if job_to_abort:
          to_abort.add(job)
    # if a job fails the whole workflow branch has to be stopped
    # look for the node in the branch to abort
    previous_size = 0
    while previous_size != len(to_abort):
      previous_size = len(to_abort)
      for dep in self.dependencies:
        job_a = self.job_mapping[dep[0]]
        job_b = self.job_mapping[dep[1]]
        if job_a in to_abort and not job_b in to_abort:
          to_abort.add(job_b)
          break

    # stop the whole branch
    ended_jobs = {}
    for job in to_abort:
      if job.job_id and job.status != constants.FAILED:
        self.logger.debug("  ---- Failure: job to abort " + job.name)
        assert(job.status == constants.NOT_SUBMITTED)
        ended_jobs[job.job_id] = job
        job.status = constants.FAILED
        job.exit_status = constants.EXIT_ABORTED


    if len(running) + len(to_run) > 0:
      status = constants.WORKFLOW_IN_PROGRESS
    elif len(done) + len(to_abort) == len(self.jobs): 
      status = constants.WORKFLOW_DONE
    elif len(done) > 0:
      status = constants.WORKFLOW_IN_PROGRESS
      # !!!! the workflow may be stuck !!!!
      # TBI
      self.logger.debug("!!!! The workflow may be stuck !!!!")
    else:
      status = constants.WORKFLOW_NOT_STARTED

    return (to_run, ended_jobs, status)

  
  def _update_state_from_database_server(self, database_server):
    wf_status = database_server.get_detailed_workflow_status(self.wf_id)

    for job_info in wf_status[0]:
      job_id, status, queue, exit_info, date_info = job_info
      self.registered_jobs[job_id].status = status
      exit_status, exit_value, term_signal, resource_usage = exit_info
      self.registered_jobs[job_id].exit_status = exit_status
      self.registered_jobs[job_id].exit_value = exit_value
      self.registered_jobs[job_id].str_rusage = resource_usage
      self.registered_jobs[job_id].terminating_signal = term_signal
   
    for ft_info in wf_status[1]:
      (engine_path, 
       client_path, 
       client_paths,
       status, 
       transfer_type) = ft_info 
      self.registered_tr[engine_path].status = status

    self.queue = wf_status[3]


  def force_stop(self, database_server):
    self._update_state_from_database_server(database_server)

    new_status = {}
    new_exit_info = {}

    self.status = constants.WORKFLOW_DONE
  
    for client_job in self.jobs:
      job = self.job_mapping[client_job]
      if not job.is_done():
        job.status = constants.FAILED
        job.exit_status = constants.EXIT_ABORTED
        job.exit_value = None
        job.terminating_signal = None
        job.drmaa_id = None
        job.str_rusage = None
        stdout = open(job.stdout_file, "w")
        stdout.close()
        stderr = open(job.stderr_file, "w")
        stderr.close()
        new_status[job.job_id] = constants.FAILED
        new_exit_info[job.job_id] = job

    database_server.set_jobs_status(new_status)
    database_server.set_jobs_exit_info(new_exit_info)
    database_server.set_workflow_status(self.wf_id, self.status)


  def restart(self, database_server, queue):
   
    self._update_state_from_database_server(database_server)

    self.queue = queue
    to_restart = False
    undone_jobs = []
    done = True
    sub_info_to_resert = {}
    new_status = {}
    jobs_queue_changed = []
    for client_job in self.jobs:
      job = self.job_mapping[client_job]
      if job.failed():
        #clear all the information related to the previous job submission
        job.status = constants.NOT_SUBMITTED
        job.exit_status = None
        job.exit_value = None
        job.terminating_signal = None
        job.drmaa_id = None
        job.queue = self.queue
        jobs_queue_changed.append(job.job_id)
        stdout = open(job.stdout_file, "w")
        stdout.close()
        stderr = open(job.stderr_file, "w")
        stderr.close()

        sub_info_to_resert[job.job_id] = None
        new_status[job.job_id] = constants.NOT_SUBMITTED

      if not job.ended_with_success():
        undone_jobs.append(job)
        job.queue = self.queue
        jobs_queue_changed.append(job.job_id)
        
    database_server.set_submission_information(sub_info_to_resert, None)
    database_server.set_jobs_status(new_status)
    database_server.set_queue(self.queue, jobs_queue_changed, self.wf_id)

    to_run = []
    if undone_jobs:
      # look for jobs to run
      for job in undone_jobs:
        job_to_run = True # a node is run when all its dependencies succeed
        for ft in job.referenced_input_files:
          eft = self.transfer_mapping[ft]
          if not eft.files_exist_on_server():
            if eft.status == constants.TRANSFERING_FROM_CR_TO_CLIENT:
              #TBI stop the transfer
              pass 
            job_to_run = False
            break 
        if job_to_run:
          for dep in self.dependencies:
            job_a = self.job_mapping[dep[0]]
            job_b = self.job_mapping[dep[1]]

            if job_b == job and not job_a.ended_with_success():
              job_to_run = False
              break

        if job_to_run: 
          to_run.append(job)
  
    if to_run:
      status = constants.WORKFLOW_IN_PROGRESS
    else:
      status = constants.WORKFLOW_DONE
     
    return (to_run, status)    



class EngineTransfer(FileTransfer):
  
  engine_path = None

  status = None

  disposal_timeout = None

  workflow_id = None

  def __init__(self, client_file_transfer):
 
    exist_on_client = client_file_transfer.initial_status == constants.FILES_ON_CLIENT
    super(EngineTransfer, self).__init__( exist_on_client,
                                          client_file_transfer.client_path,
                                          client_file_transfer.disposal_timeout,
                                          client_file_transfer.name,
                                          client_file_transfer.client_paths)

    self.status = self.initial_status

    workflow_id = -1


  def files_exist_on_server(self):

    exist = self.status == constants.FILES_ON_CR or \
            self.status == constants.FILES_ON_CLIENT_AND_CR or \
            self.status == constants.TRANSFERING_FROM_CR_TO_CLIENT
    return exist
 



# -----------------------------------------------------------
#  Copyright (C) 2008 StatPro Italia s.r.l.
#
#  StatPro Italia
#  Via G. B. Vico 4
#  I-20123 Milano
#  ITALY
#
#  phone: +39 02 96875 1
#  fax:   +39 02 96875 605
#
#  This program is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the
#  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE. See the license for more details.
# -----------------------------------------------------------
#
#  Author: Enrico Sirola <enrico.sirola@statpro.com>

"""\
A python package for DRM job submission and control.

This package is an implementation of the DRMAA 1.0 Python language
binding specification (http://www.ogf.org/documents/GFD.143.pdf).  See
http://drmaa-python.googlecode.com for package info and download.
"""

__docformat__ = "restructuredtext en"

import ctypes as _ct

try:
    import namedtuple as _nt
except ImportError: # pre 2.6 behaviour
    import nt as _nt

__version__ = "$Revision$"[11:-2]

import somadrmaa.const as _c
from somadrmaa.const import JobState, JobControlAction, JobSubmissionState
import somadrmaa.wrappers as _w
import somadrmaa.helpers as _h
from somadrmaa.errors import (AlreadyActiveSessionException,
                          AuthorizationException,
                          ConflictingAttributeValuesException,
                          DefaultContactStringException,
                          DeniedByDrmException,
                          DrmCommunicationException,
                          DrmsExitException,
                          DrmsInitException,
                          ExitTimeoutException,
                          HoldInconsistentStateException,
                          IllegalStateException,
                          InternalException,
                          InvalidAttributeFormatException,
                          InvalidContactStringException,
                          InvalidJobException,
                          InvalidJobTemplateException,
                          NoActiveSessionException,
                          NoDefaultContactStringSelectedException,
                          ReleaseInconsistentStateException,
                          ResumeInconsistentStateException,
                          SuspendInconsistentStateException,
                          TryLaterException,
                          UnsupportedAttributeException,
                          InvalidArgumentException,
                          InvalidAttributeValueException,
                          OutOfMemoryException,)

Version = _h.Version
JobInfo = _nt.namedtuple("JobInfo",
                         """jobId hasExited hasSignal terminatedSignal
                            hasCoreDump wasAborted exitStatus resourceUsage""")
# FileTransferMode = _nt.namedtuple("FileTransferMode",
#                                   """transferInputStream transferOutputStream
#                                      transferErrorStream""")

class JobTemplate(object):
    """A job to be submitted to the DRM."""

    HOME_DIRECTORY = '$drmaa_hd_ph$'
    """Home directory placeholder."""
    WORKING_DIRECTORY = '$drmaa_wd_ph$'
    """Working directory placeholder."""
    PARAMETRIC_INDEX = '$drmaa_incr_ph$'
    """Parametric index (for job arrays / bulk jobs) placeholder."""

    @property
    def attributeNames(self):
        """\
The list of supported DRMAA scalar attribute names.

This is apparently useless now, and should probably substituted by the
list of attribute names of the JobTemplate instances.
"""
        return list(_h.attribute_names_iterator())

    # scalar attributes
    remoteCommand           = _h.Attribute(_c.REMOTE_COMMAND       )
    """The command to be executed."""
    jobSubmissionState      = _h.Attribute(_c.JS_STATE             )
    """The job status."""
    workingDirectory        = _h.Attribute(_c.WD                   )
    """The job working directory."""
    jobCategory             = _h.Attribute(_c.JOB_CATEGORY         )
    """The job category."""
    nativeSpecification     = _h.Attribute(_c.NATIVE_SPECIFICATION )
    """\
A (DRM-dependant) opaque string to be passed to the DRM representing
other directives.
"""
    blockEmail              = _h.Attribute(
        _c.BLOCK_EMAIL,
        type_converter=_h.BoolConverter(true='1', false='0'))
    """False id this job should send an email, True otherwise."""
    startTime               = _h.Attribute(_c.START_TIME           )
    """The job start time, a partial timestamp string."""
    jobName                 = _h.Attribute(_c.JOB_NAME             )
    """The job Name."""
    inputPath               = _h.Attribute(_c.INPUT_PATH           )
    """The path to a file representing job's stdin."""
    outputPath              = _h.Attribute(_c.OUTPUT_PATH          )
    """The path to a file representing job's stdout."""
    errorPath               = _h.Attribute(_c.ERROR_PATH           )
    """The path to a file representing job's stderr."""
    joinFiles               = _h.Attribute(
        _c.JOIN_FILES,
        type_converter=_h.BoolConverter())
    """True if stdin and stdout should be merged, False otherwise."""
    # the following is available on ge6.2 only if enabled via cluster
    # configuration
    transferFiles           = _h.Attribute(_c.TRANSFER_FILES       )
    """\
True if file transfer should be enabled, False otherwise.

This option might require specific DRM configuration (it does on SGE).
"""
    # the following are apparently not available on ge 6.2
    # it will raise if you try to access these attrs
    deadlineTime            = _h.Attribute(_c.DEADLINE_TIME        )
    """The job deadline time, a partial timestamp string."""
    hardWallclockTimeLimit  = _h.Attribute(_c.WCT_HLIMIT, _h.IntConverter)
    """\
'Hard' Wallclock time limit, in seconds.

The job will be killed by the DRM if it takes more than
'hardWallclockTimeLimit' to complete.
"""
    softWallclockTimeLimit  = _h.Attribute(_c.WCT_SLIMIT, _h.IntConverter)
    """\
'Soft' Wallclock time limit, in seconds.

The job will be signaled by the DRM if it takes more than
'hardWallclockTimeLimit' to complete.
"""
    hardRunDurationLimit    = _h.Attribute(_c.DURATION_HLIMIT, _h.IntConverter)
    """\
WRITE ME.
"""
    softRunDurationLimit    = _h.Attribute(_c.DURATION_SLIMIT, _h.IntConverter)
    """\
WRITE ME.
"""

    # vector attributes
    email = _h.VectorAttribute(_c.V_EMAIL)
    """email addresses to whom send job completion info."""
    args  = _h.VectorAttribute(_c.V_ARGV)
    """The job's command argument list."""
    # dict attributes
    jobEnvironment = _h.DictAttribute(_c.V_ENV)
    """The job's environment dict."""

    _as_parameter_ = None

    def __init__(self, **kwargs):
        """\
Builds a JobTemplate instance.

Attributes can be passed as keyword arguments."""
        jt = _ct.pointer(_ct.POINTER(_w.drmaa_job_template_t)())
        _h.c(_w.drmaa_allocate_job_template, jt)
        self._jt = self._as_parameter_ = jt.contents
        try:
            for aname in kwargs:
                setattr(self, aname, kwargs.get(aname))
        except:
            self.delete()
            raise

    def delete(self):
        """Deallocate the underlying DRMAA job template."""
        _h.c(_w.drmaa_delete_job_template, self)

    def __enter__(self):
        """context manager enter routine"""
        return self

    def __exit__(self, *_):
        """\
context manager exit routine.

Stops communication with the DRM.
"""
        self.delete()
        return False

class Session(object):
    """\
The DRMAA Session.

This class is the entry point for communicating with the DRM system
"""
    TIMEOUT_WAIT_FOREVER = _c.TIMEOUT_WAIT_FOREVER
    TIMEOUT_NO_WAIT = _c.TIMEOUT_NO_WAIT
    JOB_IDS_SESSION_ANY = _c.JOB_IDS_SESSION_ANY
    JOB_IDS_SESSION_ALL = _c.JOB_IDS_SESSION_ALL

    contact = _h.SessionStringAttribute(_w.drmaa_get_contact)
    """\
a comma delimited string list containing the contact strings available
from the default DRMAA implementation, one element per DRM system
available. If called after initialize(), this method returns the
contact String for the DRM system to which the session is
attached. The returned strings are implementation dependent.
"""

    drmsInfo = _h.SessionStringAttribute(_w.drmaa_get_DRM_system)
    """\
If called before initialize(), this method returns a comma delimited
list of DRM systems, one element per DRM system implementation
provided. If called after initialize(), this method returns the
selected DRM system. The returned String is implementation dependent.
"""
    drmaaImplementation = _h.SessionStringAttribute(
        _w.drmaa_get_DRMAA_implementation)
    """\
If called before initialize(), this method returns a comma delimited
list of DRMAA implementations, one element for each DRMAA
implementation provided. If called after initialize(), this method
returns the selected DRMAA implementation. The returned String is
implementation dependent and may contain the DRM system as a
component.
"""
    version = _h.SessionVersionAttribute()
    """\
a Version object containing the major and minor version numbers of the
DRMAA library. For DRMAA 1.0, major is 1 and minor is 0.
"""
    def __init__(self, contactString=None):
        self.contactString = contactString

    # no return value
    @staticmethod
    def initialize(contactString=None):
        """\
Used to initialize a DRMAA session for use.

:Parameters:
  contactString : string or None
     implementation-dependent string that
     may be used to specify which DRM system to use

This method must be called before any other DRMAA calls.  If
contactString is None, the default DRM system is used, provided there
is only one DRMAA implementation available. If there is more than one
DRMAA implementation available, initialize() throws a
NoDefaultContactStringSelectedException. initialize() should be called
only once, by only one of the threads. The main thread is
recommended. A call to initialize() by another thread or additional
calls to initialize() by the same thread with throw a
SessionAlreadyActiveException.
"""
        _w.init(contactString)

    # no return value
    @staticmethod
    def exit():
        """\
Used to disengage from DRM.

This routine ends the current DRMAA session but doesn't affect any
jobs (e.g., queued and running jobs remain queued and
running). exit() should be called only once, by only one of the
threads. Additional calls to exit() beyond the first will throw a
NoActiveSessionException.
"""
        _w.exit()

    # returns JobTemplate instance
    @staticmethod
    def createJobTemplate():
        """\
Allocates a new job template.

The job template is used to set the environment for jobs to be
submitted. Once the job template has been created, it should also be
deleted (via deleteJobTemplate()) when no longer needed. Failure to do
so may result in a memory leak.
"""
        return JobTemplate()

    # takes JobTemplate instance, no return value
    @staticmethod
    def deleteJobTemplate(jobTemplate):
        """\
Deallocate a job template.

:Parameters:
  jobTemplate : JobTemplate
    the job temptare to be deleted

This routine has no effect on running jobs.
"""
        jobTemplate.delete()

    # takes JobTemplate instance, returns string
    @staticmethod
    def runJob(jobTemplate):
        """\
Submit a job with attributes defined in the job template.

:Parameters:
  jobTemplate : JobTemplate
    the template representing the job to be run

The returned job identifier is a String identical to that returned
from the underlying DRM system.
"""
        jid = _ct.create_string_buffer(128)
        _h.c(_w.drmaa_run_job, jid, _ct.sizeof(jid), jobTemplate)
        return jid.value

    # takes JobTemplate instance and num values, returns string list
    @staticmethod
    def runBulkJobs(jobTemplate, beginIndex, endIndex, step):
        """\
Submit a set of parametric jobs, each with attributes defined in the job template.

:Parameters:
  jobTemplate : JobTemplate
    the template representng jobs to be run
  beginIndex : int
    index of the first job
  endIndex : int
    index of the last job
  step : int
    the step between job ids

The returned job identifiers are Strings identical to those returned
from the underlying DRM system.  The JobTemplate class defines a
`JobTemplate.PARAMETRIC_INDEX` placeholder for use in specifying paths.
This placeholder is used to represent the individual identifiers of
the tasks submitted through this method.
"""
        return list(_h.run_bulk_job(jobTemplate, beginIndex, endIndex, step))

    # takes string and JobControlAction value, no return value
    @staticmethod
    def control(jobId, operation):
        """\
Used to hold, release, suspend, resume, or kill the job identified by jobId.  

:Parameters:
  jobId : string
    if jobId is `Session.JOB_IDS_SESSION_ALL` then this routine acts on
    all jobs submitted during this DRMAA session up to the moment
    control() is called. The legal values for
    action and their meanings are
  operation : string
    possible values are:
        `JobControlAction.SUSPEND`
          stop the job
        `JobControlAction.RESUME`
          (re)start the job
        `JobControlAction.HOLD`
          put the job on-hold
        `JobControlAction.RELEASE`
          release the hold on the job
        `JobControlAction.TERMINATE`
          kill the job

To avoid thread races in multithreaded applications, the DRMAA
implementation user should explicitly synchronize this call with
any other job submission calls or control calls that may change
the number of remote jobs.

This method returns once the action has been acknowledged by the DRM
system, but does not necessarily wait until the action has been
completed.  Some DRMAA implementations may allow this method to be
used to control jobs submitted external to the DRMAA session, such as
jobs submitted by other DRMAA session in other DRMAA implementations
or jobs submitted via native utilities.
"""
        _h.c(_w.drmaa_control, jobId, _c.string_to_control_action(operation))

    # takes string list, num value and boolean, no return value
    @staticmethod
    def synchronize(jobIds, timeout=-1, dispose=False):
        """\
Waits until all jobs specified by jobList have finished execution.

:Parameters:
  jobIds
    If jobIds contains `Session.JOB_IDS_SESSION_ALL`, then this
    method waits for all jobs submitted during this DRMAA session up to
    the moment synchronize() is called
  timeout : int
    maximum time (in seconds) to be waited for the completion of a job.

    The value `Session.TIMEOUT_WAIT_FOREVER` may be specified to wait
    indefinitely for a result. The value `Session.TIMEOUT_NO_WAIT` may
    be specified to return immediately if no result is available.
  dispose : bool
    specifies how to treat the reaping of the remote job's internal
    data record, which includes a record of the job's consumption of
    system resources during its execution and other statistical
    information. If set to True, the DRM will dispose of the job's
    data record at the end of the synchroniize() call. If set to
    False, the data record will be left for future access via the
    wait() method.

To avoid thread race conditions in multithreaded applications, the
DRMAA implementation user should explicitly synchronize this call with
any other job submission calls or control calls that may change the
number of remote jobs.

If the call exits before the
timeout has elapsed, all the jobs have been waited on or there was an
interrupt. If the invocation exits on timeout, an ExitTimeoutException
is thrown. The caller should check system time before and after this
call in order to be sure of how much time has passed.  
"""
        if dispose: d = 1
        else: d = 0
        _h.c(_w.drmaa_synchronize, _h.string_vector(jobIds), timeout, d)

    # takes string and long, returns JobInfo instance
    @staticmethod
    def wait(jobId, timeout=-1):
        """\
Wait for a job with jobId to finish execution or fail. 

:Parameters:
  `jobId` : str
    The job id to wait completion for.

    If the special string, `Session.JOB_IDS_SESSION_ANY`, is provided as the
    jobId, this routine will wait for any job from the session
  `timeout` : float
    The timeout value is used to specify the desired behavior when a
    result is not immediately available. 

    The value `Session.TIMEOUT_WAIT_FOREVER` may be specified to wait
    indefinitely for a result. The value `Session.TIMEOUT_NO_WAIT` may
    be specified to return immediately if no result is
    available. Alternatively, a number of seconds may be specified to
    indicate how long to wait for a result to become available

This routine is modeled on the wait3 POSIX routine. If the call exits
before timeout, either the job has been waited on successfully or
there was an interrupt. If the invocation exits on timeout, an
`ExitTimeoutException` is thrown. The caller should check system time
before and after this call in order to be sure how much time has
passed.  The routine reaps job data records on a successful call, so
any subsequent calls to wait() will fail, throwing an
`InvalidJobException`, meaning that the job's data record has been
already reaped.  This exception is the same as if the job were
unknown. (The only case where wait() can be successfully called on a
single job more than once is when the previous call to wait() timed
out before the job finished.)
"""
        stat = _ct.c_int()
        jid_out = _ct.create_string_buffer(128)
        job_id_out = _ct.c_int()
        rusage = _ct.pointer(_ct.POINTER(_w.drmaa_attr_values_t)())
        _h.c(_w.drmaa_wait, jobId, jid_out, _ct.sizeof(jid_out),
             _ct.byref(stat), timeout, rusage)
        res_usage = _h.adapt_rusage(rusage)
        exited = _ct.c_int()
        _h.c(_w.drmaa_wifexited, _ct.byref(exited), stat)
        aborted = _ct.c_int()
        _h.c(_w.drmaa_wifaborted, _ct.byref(aborted), stat)
        signaled = _ct.c_int()
        _h.c(_w.drmaa_wifsignaled, _ct.byref(signaled), stat)
        coredumped = _ct.c_int()
        _h.c(_w.drmaa_wcoredump, _ct.byref(coredumped), stat)
        exit_status = _ct.c_int()
        _h.c(_w.drmaa_wexitstatus, _ct.byref(exit_status), stat)
        term_signal = _ct.create_string_buffer(_c.SIGNAL_BUFFER)
        _h.c(_w.drmaa_wtermsig, term_signal, _ct.sizeof(term_signal), stat)
        return JobInfo(jid_out.value, bool(exited), bool(signaled),
                       term_signal.value, bool(coredumped),
                       bool(aborted), int(exit_status.value), res_usage)

    # takes string, returns JobState instance
    @staticmethod
    def jobStatus(jobName):
        """\
returns the program status of the job identified by jobId.

The possible values returned from
this method are:

* `JobState.UNDETERMINED`: process status cannot be determined,
* `JobState.QUEUED_ACTIVE`: job is queued and active,
* `JobState.SYSTEM_ON_HOLD`: job is queued and in system hold,
* `JobState.USER_ON_HOLD`: job is queued and in user hold,
* `JobState.USER_SYSTEM_ON_HOLD`: job is queued and in user and system hold,
* `JobState.RUNNING`: job is running,
* `JobState.SYSTEM_SUSPENDED`: job is system suspended,
* `JobState.USER_SUSPENDED`: job is user suspended,
* `JobState.DONE`: job finished normally, and
* `JobState.FAILED`: job finished, but failed.

The DRMAA implementation should always get the status of the job from
the DRM system unless the status has already been determined to be
FAILED or DONE and the status has been successfully cached. Terminated
jobs return a FAILED status.
"""
        status = _ct.c_int()
        _h.c(_w.drmaa_job_ps, jobName, _ct.byref(status))
        return _c.status_to_string(status.value)

    def __enter__(self):
        """Context manager enter function"""
        self.initialize(self.contactString)
        return self

    def __exit__(self, *_):
        """Context manager exit function."""
        self.exit()
        return False



"""
Soma-workflow is a unified and simple interface to parallel computing
resources. It aims at making easier the use of parallel resources by
non expert users and software.
"""
import ez_setup
ez_setup.use_setuptools()
from setuptools import setup
import os.path as op
commands = [op.join('bin', 'soma_delete_all_workflows'),
            op.join('bin', 'soma_stop_workflow'),
            op.join('bin', 'soma_workflow_gui'),
            op.join('bin', 'soma_restart_workflow'),
            op.join('bin', 'soma_submit_workflow'), ]
setup(
    name="soma-workflow",
    description=__doc__,
    long_description=file('README.txt', 'r').read(),
    license='CeCILL v2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        ' Environment :: X11 Applications ',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Distributed Computing',
    ],
    author='Please check authors at http://www.brainvisa.info/soma-workflow',
    author_email='jinpeng.li@cea.fr',
    version='2.5.2',
    url='http://www.brainvisa.info/soma-workflow',
    package_dir={'': 'python'},
    packages=['somadrmaa',
              'soma_workflow',
              'soma_workflow.gui',
              'soma_workflow.test',
              'soma_workflow.check_requirement'],
    include_package_data=True,
    package_data={'soma_workflow.gui': ['*ui', 'icon/*png']},
    scripts=commands,
    platforms=['Linux', 'MacOS X', 'Windows'],
    # requires=['PyQt', ],
    extras_require={
        'plotting': ['matplotlib'],
        'client': ['Pyro', 'paramiko'],
    },
)
